use clap::{App, Arg};
use fat32::{Attribute, BlockDevice, FAT32Manager, Inode};
use std::fs::{read_dir, File, OpenOptions};
use std::io::{Read, Seek, SeekFrom, Write};
use std::sync::Arc;
use std::sync::Mutex;

const BLOCK_SZ: usize = 512;

struct BlockFile(Mutex<File>);

impl BlockDevice for BlockFile {
    fn read_block(&self, block_id: usize, buf: &mut [u8]) {
        let mut file = self.0.lock().unwrap();
        file.seek(SeekFrom::Start((block_id * BLOCK_SZ) as u64))
            .expect("Error when seeking!");
        assert_eq!(file.read(buf).unwrap(), BLOCK_SZ, "Not a complete block!");
    }

    fn write_block(&self, block_id: usize, buf: &[u8]) {
        let mut file = self.0.lock().unwrap();
        file.seek(SeekFrom::Start((block_id * BLOCK_SZ) as u64))
            .expect("Error when seeking!");
        assert_eq!(file.write(buf).unwrap(), BLOCK_SZ, "Not a complete block!");
    }
}

fn main() {
    fat32_pack().expect("Error when packing fat32!");
}

fn fat32_pack() -> std::io::Result<()> {
    // clap::matches 用于捕获用户输入的参数
    // 在makefile中，命令为
    // @cd ../easy-fs-fuse && cargo run --release \
    // -- -s ../user/src/bin/ \
    // -t ../user/target/riscv64gc-unknown-none-elf/release/
    // 因此得到的参数就是两个路径
    let matches = App::new("EasyFileSystem packer")
        .arg(
            Arg::with_name("source")
                .short("s") // 对应输入的 -s
                .long("source") //对应输入 --source
                .takes_value(true)
                .help("Executable source dir(with backslash)"),
        )
        .arg(
            Arg::with_name("target")
                .short("t")
                .long("target")
                .takes_value(true)
                .help("Executable target dir(with backslash)"),
        )
        .get_matches();
    let src_path = matches.value_of("source").unwrap();
    let target_path = matches.value_of("target").unwrap();
    println!("src_path = {}\ntarget_path = {}", src_path, target_path);

    // 打开U盘
    let block_file = Arc::new(BlockFile(Mutex::new({
        let f = OpenOptions::new()
            .read(true)
            .write(true)
            .create(true)
            // .open("/dev/sdb")?;
            .open("fat32.img")?;
        // 注意当新建文件的时候大小是0，需要先设置大小，这样读块的时候不会出错
        f.set_len(8192 * 512).unwrap();
        f
    })));

    let fs = FAT32Manager::open(block_file.clone());
    let mut root_inode = FAT32Manager::root_inode(&fs);

    // 从host获取应用名
    let apps: Vec<_> = read_dir(src_path)
        .unwrap()
        .into_iter()
        .map(|dir_entry| {
            let mut name_with_ext = dir_entry.unwrap().file_name().into_string().unwrap();
            // 丢弃后缀 从'.'到末尾(len-1)
            name_with_ext.drain(name_with_ext.find('.').unwrap()..name_with_ext.len());
            name_with_ext
        })
        .collect();
    for app in apps {
        // load app data from host file system
        let mut host_file = File::open(format!("{}{}", target_path, app)).unwrap();
        let mut all_data: Vec<u8> = Vec::new();
        host_file.read_to_end(&mut all_data).unwrap();
        // create a file in easy-fs
        println!("before create");
        let o_inode = root_inode.create(app.as_str(), Attribute::ARCHIVE);
        if o_inode.is_none() {
            continue;
        }
        let mut inode = o_inode.unwrap();
        println!("after create");
        // write data to easy-fs
        println!("file_len = {}", all_data.len());
        inode.write_at(0, all_data.as_slice());
    }

    // list apps
    for app in root_inode.ls() {
        println!("{}", app);
    }
    Ok(())
}

#[allow(unused)]
macro_rules! color_text {
    ($text:expr, $color:expr) => {{
        format_args!("\x1b[{}m{}\x1b[0m", $color, $text)
    }};
}

#[cfg(test)]
mod tests {
    use super::*;

    fn ls(inode: &Inode) {
        println!("----------------- ls -----------------");
        for name in inode.ls() {
            println!("{} ", name);
        }
        println!("--------------------------------------\n");
    }

    fn simple_rwtest(inode: &mut Inode) {
        let greet_str = "hello world!";
        println!("\n----------------- simple r/w test -----------------");
        println!(
            "  1: write file. wlen={}",
            inode.write_at(0, greet_str.as_bytes())
        );
        let mut buffer = [0u8; 256];
        let len = inode.read_at(0, &mut buffer);
        println!("  2: read file. rlen = {}", len);
        println!("  text = {}", core::str::from_utf8(&buffer[..len]).unwrap());
        assert_eq!(greet_str, core::str::from_utf8(&buffer[..len]).unwrap(),);
        println!("----------------- simple r/w test pass -----------------\n");
    }

    #[test]
    fn ufs_test() -> std::io::Result<()> {
        let block_file = Arc::new(BlockFile(Mutex::new({
            let f = OpenOptions::new()
                .read(true)
                .write(true)
                .create(true)
                .open("./fat32.img")?;
            f
        })));

        let fs = FAT32Manager::open(block_file.clone());

        // ----------------------------------------
        // 测试根目录
        // ----------------------------------------
        println!("\n================= 测试根目录 =====================");
        let mut root_inode = FAT32Manager::root_inode(&fs);
        println!("root name = {}", root_inode.name());
        println!("pwd = {}", root_inode.path());
        ls(&root_inode);
        println!("\n=============================================\n");


        // ----------------------------------------
        // 测试创建文件
        // ----------------------------------------
        println!("\n================ 测试创建文件 ===================");
        // 创建短文件
        root_inode.create("hello", Attribute::ARCHIVE).unwrap();
        let mut hello = root_inode.namei("hello").unwrap();
        println!("create file: name = {}", hello.name());
        simple_rwtest(&mut hello);
        ls(&root_inode);

        // 创建长文件
        root_inode.create("longname_longname_longname_longname", Attribute::ARCHIVE).unwrap();
        let mut long_name = root_inode.namei("longname_longname_longname_longname").unwrap();
        println!("create file: name = {}", long_name.name());
        simple_rwtest(&mut long_name);
        ls(&root_inode);

        // 创建重复文件
        assert!(root_inode.create("longname_longname_longname_longname", Attribute::ARCHIVE).is_none());
        println!("\n=============================================\n");


        // ----------------------------------------
        // 测试目录文件
        // ----------------------------------------
        println!("\n================= 测试目录文件 ==================");
        println!("directory test ... start");
        let mut dir0 = root_inode.create("dir0", Attribute::DIRECTORY).unwrap();
        ls(&root_inode);

        println!("list dir0:");
        ls(&dir0);

        dir0.create("file1", Attribute::ARCHIVE).unwrap();
        let mut dir0_dir1 = dir0.create("dir1", Attribute::DIRECTORY).unwrap();
        dir0_dir1.create("file2", Attribute::ARCHIVE);

        let mut file1 = root_inode.namei("dir0/file1").unwrap();
        let mut file2 = root_inode.namei("dir0/dir1/file2/").unwrap();
        simple_rwtest(&mut file1);
        simple_rwtest(&mut file2);
        println!("directory test ... end");
        println!("\n=============================================\n");


        // ----------------------------------------
        // 测试长复杂路径
        // ----------------------------------------
        println!("\n================= 测试长复杂路径 ==================");
        let mut aaa = root_inode.create("aaa", Attribute::DIRECTORY).unwrap();
        let mut bbb = aaa.create("bbb", Attribute::DIRECTORY).unwrap();
        let mut ccc = bbb.create("ccc", Attribute::DIRECTORY).unwrap();
        ccc.create("ddd", Attribute::ARCHIVE).unwrap();

        let file = bbb.namei("../.").unwrap();
        assert_eq!("AAA", file.name());
        println!("open file name = {}", file.name());
        println!("pwd = {}", file.path());

        println!("list bbb:");
        ls(&bbb);

        let mut file = bbb.namei("../../../aaa/bbb/./../bbb/ccc/ddd").unwrap();
        assert_eq!("DDD", file.name());
        println!("open file name = {}", file.name());
        println!("pwd = {}", file.path());
        simple_rwtest(&mut file);
        println!("\n=============================================\n");


        // ----------------------------------------
        // 测试随机读写
        // ----------------------------------------
        println!("\n================ 测试随机读写 ===================");
        println!("random str rw test ... start");
        let mut buffer = [0u8; 256];
        let mut filea = root_inode.create("filea", Attribute::ARCHIVE).unwrap();
        let mut random_str_test = |len: usize| {
            filea.clear();
            assert_eq!(filea.read_at(0, &mut buffer), 0,);
            let mut str = String::new();
            use rand;
            // random digit
            for _ in 0..len {
                str.push(char::from('0' as u8 + rand::random::<u8>() % 10));
            }
            let mut str1 = str.clone();
            filea.write_at(0, str.as_bytes());
            println!("after write, size = {}", filea.size());
            unsafe {
                filea.read_at(0, str1.as_bytes_mut());
            }
            assert_eq!(str, str1);
            drop(str1);
            let mut read_buffer = [0u8; 127];
            let mut offset = 0usize;
            let mut read_str = String::new();
            loop {
                let len = filea.read_at(offset, &mut read_buffer);
                if len == 0 {
                    break;
                }
                offset += len;
                read_str.push_str(core::str::from_utf8(&read_buffer[..len]).unwrap());
            }
            println!("offset = {} ... pass", offset);
            assert_eq!(str.len(), read_str.len());
        };

        random_str_test(4 * BLOCK_SZ);
        random_str_test(8 * BLOCK_SZ + BLOCK_SZ / 2);
        random_str_test(33 * BLOCK_SZ);
        random_str_test(70 * BLOCK_SZ + BLOCK_SZ / 7);
        random_str_test((12 + 128) * BLOCK_SZ);
        random_str_test(400 * BLOCK_SZ);
        random_str_test(1000 * BLOCK_SZ);
        random_str_test(2000 * BLOCK_SZ);

        println!("random str rw test ... {}", color_text!("pass", 92));
        println!("\n=============================================\n");

        Ok(())
    }
}
