use crate::END_CLUSTER;
use alloc::string::String;

pub const DIRENT_SZ: usize = 32;
pub const SHORT_PRIMARY_LEN: usize = 8;
pub const SHORT_EXT_LEN: usize = 3;
pub const LONG_NAME_LEN: u32 = 13;

// 显示规则
pub const ALL_UPPER_CASE: u8 = 0x00;
pub const NAME_LOWER_EXT_UPPER: u8 = 0x08;
pub const NAME_UPPER_EXT_LOWER: u8 = 0x10;
pub const ALL_LOWER_CASE: u8 = 0x18;

/// 目录项结构与实际存储设备一致，便于读写
/// repr(packed) 限制结构体非对齐
/// 没有时间处理
#[derive(Clone, Copy, Debug)]
#[repr(packed)]
pub struct ShortDirEntry {
    name: [u8; 8], // 删除时第0位为0xE5，未使用时为0x00. 有多余可以用0x20填充
    extension: [u8; 3],
    pub attribute: u8,
    winnt_reserved: u8,
    _creation_tenths: u8, // 创建时间的 10 毫秒位
    creation_time: u16,
    creation_date: u16,
    last_acc_date: u16,
    cluster_high: u16,
    modification_time: u16,
    modification_date: u16,
    cluster_low: u16,
    pub size: u32,
}

impl ShortDirEntry {
    /// 空短目录项
    pub fn empty() -> Self {
        unsafe { *(&[0; 0x20] as *const _ as usize as *const Self) }
    }

    /// 新建短目录项
    /// 
    /// 对于新创建的目录文件分配一个簇的空间，其它文件不分配
    pub fn new(name: [u8; 8], extension: [u8; 3], attribute: u8, cluster: u32) -> Self {
        let mut res = Self {
            name,
            extension,
            attribute,
            winnt_reserved: 0,
            _creation_tenths: 0,
            creation_time: 0,
            creation_date: 0x54d9,
            last_acc_date: 0,
            cluster_high: 0xfff,
            modification_time: 0,
            modification_date: 0,
            cluster_low: 0xffff,
            size: 0,
        };
        res.set_first_cluster(cluster);
        res
    }

    pub fn is_deleted(&self) -> bool {
        self.name[0] == 0xE5
    }

    /// 0 表示该目录项是空闲的，同时表明该目录项之后的所有目录项也是空闲的，
    /// 可以直接使用
    pub fn is_free(&self) -> bool {
        self.name[0] == 0x00
    }

    pub fn is_valid(&self) -> bool {
        !self.is_deleted() && !self.is_free()
    }

    pub fn is_dir(&self) -> bool {
        Attribute::from_bits(self.attribute)
            .unwrap()
            .contains(Attribute::DIRECTORY)
    }

    pub fn is_file(&self) -> bool {
        !self.is_dir()
    }

    pub fn is_short(&self) -> bool {
        !Attribute::from_bits(self.attribute)
            .unwrap()
            .contains(Attribute::LONG_NAME)
    }

    pub fn get_creation_time(&self) -> (u32, u32, u32, u32, u32, u32, u64) {
        // year-month-day-Hour-min-sec-long_sec
        let year: u32 = ((self.creation_date & 0xFE00) >> 9) as u32 + 1980;
        let month: u32 = ((self.creation_date & 0x01E0) >> 5) as u32;
        let day: u32 = (self.creation_date & 0x001F) as u32;
        let hour: u32 = ((self.creation_time & 0xF800) >> 11) as u32;
        let min: u32 = ((self.creation_time & 0x07E0) >> 5) as u32;
        let sec: u32 = ((self.creation_time & 0x001F) << 1) as u32; // 秒数需要*2
        let long_sec: u64 =
            ((((year - 1970) * 365 + month * 30 + day) * 24 + hour) * 3600 + min * 60 + sec) as u64;
        (year, month, day, hour, min, sec, long_sec)
    }

    pub fn get_modification_time(&self) -> (u32, u32, u32, u32, u32, u32, u64) {
        // year-month-day-Hour-min-sec
        let year: u32 = ((self.modification_date & 0xFE00) >> 9) as u32 + 1980;
        let month: u32 = ((self.modification_date & 0x01E0) >> 5) as u32;
        let day: u32 = (self.modification_date & 0x001F) as u32;
        let hour: u32 = ((self.modification_time & 0xF800) >> 11) as u32;
        let min: u32 = ((self.modification_time & 0x07E0) >> 5) as u32;
        let sec: u32 = ((self.modification_time & 0x001F) << 1) as u32; // 秒数需要*2
        let long_sec: u64 =
            ((((year - 1970) * 365 + month * 30 + day) * 24 + hour) * 3600 + min * 60 + sec) as u64;
        (year, month, day, hour, min, sec, long_sec)
    }

    pub fn get_accessed_time(&self) -> (u32, u32, u32, u32, u32, u32, u64) {
        // year-month-day-Hour-min-sec
        let year: u32 = ((self.last_acc_date & 0xFE00) >> 9) as u32 + 1980;
        let month: u32 = ((self.last_acc_date & 0x01E0) >> 5) as u32;
        let day: u32 = (self.last_acc_date & 0x001F) as u32;
        let hour: u32 = 0;
        let min: u32 = 0;
        let sec: u32 = 0; // 没有相关信息，默认0
        let long_sec: u64 =
            ((((year - 1970) * 365 + month * 30 + day) * 24 + hour) * 3600 + min * 60 + sec) as u64;
        (year, month, day, hour, min, sec, long_sec)
    }

    /// 获取文件数据区第一个簇号</br>
    /// 注：有些代表根目录的目录项 (`.`, `..`) 其第一个簇号设置为 0（文档要求）</br>
    /// 但好像我的实现没有这样设置，但我们从挂载得到的文件系统镜像就不一定了，这里还是得转换
    pub fn first_cluster(&self) -> u32 {
        match ((self.cluster_high as u32) << 16) + (self.cluster_low as u32) {
            0 => 2,
            first_cluster => first_cluster,
        }
    }

    /// 获取短文件名
    /// 短文件名都以大写形式存储，因此返回的字符串为大写
    pub fn name(&self) -> String {
        let mut name: String = String::new();
        for i in 0..8 {
            // 记录文件名
            if self.name[i] == 0x20 {
                break;
            } else {
                name.push(self.name[i] as char);
            }
        }
        for i in 0..3 {
            // 记录扩展名
            if self.extension[i] == 0x20 {
                break;
            } else {
                if i == 0 {
                    name.push('.');
                }
                name.push(self.extension[i] as char);
            }
        }
        name
    }

    /// 计算校验和
    ///
    /// 由于 rust 对算术运算溢出 panic，因此这里使用更大的类型，然后每次对高位清零
    pub fn checksum(&self) -> u8 {
        let mut name_buff: [u8; 11] = [0u8; 11];
        let mut sum: usize = 0;
        for i in 0..8 {
            name_buff[i] = self.name[i];
        }
        for i in 0..3 {
            name_buff[i + 8] = self.extension[i];
        }
        for i in 0..11 {
            if (sum & 1) != 0 {
                sum = 0x80 + (sum >> 1) + name_buff[i] as usize;
            } else {
                sum = (sum >> 1) + name_buff[i] as usize;
            }
            sum &= u8::MAX as usize;
        }
        sum as u8
    }

    /* 设置当前文件的大小 */
    // 簇的分配和回收实际要对FAT表操作
    pub fn set_size(&mut self, size: u32) {
        self.size = size;
    }

    pub fn set_case(&mut self, case: u8) {
        self.winnt_reserved = case;
    }

    /* 设置文件起始簇 */
    pub fn set_first_cluster(&mut self, cluster: u32) {
        self.cluster_high = ((cluster & 0xFFFF0000) >> 16) as u16;
        self.cluster_low = (cluster & 0x0000FFFF) as u16;
    }

    /// 清空文件
    pub fn clear(&mut self) {
        self.size = 0;
        self.set_first_cluster(END_CLUSTER);
    }

    pub fn delete(&mut self) {
        self.size = 0;
        self.name[0] = 0xE5;
        self.set_first_cluster(END_CLUSTER);
    }

    pub fn as_bytes(&self) -> &[u8] {
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, DIRENT_SZ) }
    }
    pub fn as_bytes_mut(&mut self) -> &mut [u8] {
        unsafe { core::slice::from_raw_parts_mut(self as *mut _ as usize as *mut u8, DIRENT_SZ) }
    }
}

#[repr(packed)]
#[derive(Clone, Copy, Debug)]
pub struct LongDirEntry {
    // use Unicode !!!
    // 如果是该文件的最后一个长文件名目录项，
    // 则将该目录项的序号与 0x40 进行“或（OR）运算”的结果写入该位置。
    // 长文件名以 \0 结尾，其它位置填充 0xff
    pub order: u8,     // 删除时为0xE5
    name1: [u8; 10],   // 5characters
    pub attribute: u8, // should be 0x0F
    _reserved: u8,
    pub checksum: u8,
    name2: [u8; 12], // 6characters
    _zero: [u8; 2],
    name3: [u8; 4], // 2characters
}

impl LongDirEntry {
    pub fn empty() -> Self {
        unsafe { *(&[0; 0x20] as *const _ as usize as *const Self) }
    }

    /// 上层要完成对 namebuffer 的填充，注意末尾添加 \0，以及checksum的计算
    /// 目前只支持英文，因此传入ascii
    pub fn new(name_buffer: &[u8; 13], order: u8, checksum: u8) -> Self {
        let mut name: [u8; 26] = [0xff; 26];
        for i in 0..13 {
            name[i << 1] = name_buffer[i];
            name[(i << 1) + 1] = 0;
            if name_buffer[i] == 0 {
                break;
            }
        }
        Self {
            order,
            name1: (&name[0..10]).try_into().unwrap(),
            attribute: Attribute::LONG_NAME.bits(),
            _reserved: 0,
            checksum,
            name2: (&name[10..22]).try_into().unwrap(),
            _zero: [0u8; 2],
            name3: (&name[22..26]).try_into().unwrap(),
        }
    }

    pub fn is_free(&self) -> bool {
        self.order == 0
    }

    pub fn is_deleted(&self) -> bool {
        self.order == 0xE5
    }

    pub fn is_valid(&self) -> bool {
        !self.is_deleted() && !self.is_free()
    }

    pub fn delete(&mut self) {
        self.order = 0xE5;
    }

    // 获取长文件名，此处完成unicode至ascii的转换，暂不支持中文！
    // 这里直接将几个字段拼合，不考虑填充字符0xFF等
    // 需要和manager的long_name_split配合使用
    pub fn get_name_raw(&self) -> String {
        let mut name = String::new();
        for i in 0..5 {
            name.push(self.name1[i << 1] as char);
        }
        for i in 0..6 {
            name.push(self.name2[i << 1] as char);
        }
        for i in 0..2 {
            name.push(self.name3[i << 1] as char);
        }
        return name;
    }

    pub fn get_name_format(&self) -> String {
        let mut name = String::new();
        let mut c: u8;
        for i in 0..5 {
            c = self.name1[i << 1];
            if c == 0 {
                return name;
            }
            name.push(c as char);
        }
        for i in 0..6 {
            c = self.name2[i << 1];
            if c == 0 {
                return name;
            }
            name.push(c as char);
        }
        for i in 0..2 {
            c = self.name3[i << 1];
            if c == 0 {
                return name;
            }
            name.push(c as char);
        }
        return name;
    }

    pub fn as_bytes(&self) -> &[u8] {
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, DIRENT_SZ) }
    }
    pub fn as_bytes_mut(&mut self) -> &mut [u8] {
        unsafe { core::slice::from_raw_parts_mut(self as *mut _ as usize as *mut u8, DIRENT_SZ) }
    }
}

bitflags! {
    pub struct Attribute: u8 {
        const RDONLY = 0x01;
        const HIDDEN = 0x02;
        const SYSTEM = 0x04;
        const VOLUME_ID = 0x08;
        const DIRECTORY = 0x10;
        const ARCHIVE = 0x20;
        const LONG_NAME = Self::RDONLY.bits
                        | Self::HIDDEN.bits
                        | Self::SYSTEM.bits
                        | Self::VOLUME_ID.bits;
    }
}
