use super::{get_block_cache, BlockDevice, BLOCK_SZ};
use alloc::sync::Arc;
use alloc::vec::Vec;

const LEAD_SIGNATURE: u32 = 0x41615252;
const SECOND_SIGNATURE: u32 = 0x61417272;
pub const FREE_CLUSTER: u32 = 0x00000000;
pub const BOUND_CLUSTER: u32 = 0x0FFFFFF8;
pub const END_CLUSTER: u32 = 0x0FFFFFFF;
pub const BAD_CLUSTER: u32 = 0x0FFFFFF7;
const FATENTRY_PER_SEC: u32 = BLOCK_SZ as u32 / 4;

type DataBlock = [u8; BLOCK_SZ];

#[repr(packed)]
#[derive(Clone, Copy, Debug)]
pub struct BPB {
    pub bytes_per_sector: u16,   // 每扇区字节数
    pub sectors_per_cluster: u8, // 每簇扇区数
    pub reserved_sectors: u16,   // 保留扇区数
    pub fats: u8,                // FAT 表数目
    pub total_sectors: u32,      // 总扇区数
    pub sectors_per_fat: u32,    // 一个 FAT 表扇区数
    pub root_clusters: u32,      // 第一个目录簇号
    pub fsinfo_sector: u16,      // fsinfo扇区的扇区号
}

impl BPB {
    pub fn new(dbr: &DataBlock) -> Self {
        unsafe {
            Self {
                bytes_per_sector: *(dbr.as_ptr().add(0xb) as *const u16),
                sectors_per_cluster: *(dbr.as_ptr().add(0xd)),
                reserved_sectors: *(dbr.as_ptr().add(0xe) as *const u16),
                fats: *(dbr.as_ptr().add(0x10)),
                total_sectors: *(dbr.as_ptr().add(0x20) as *const u32),
                sectors_per_fat: *(dbr.as_ptr().add(0x24) as *const u32),
                root_clusters: *(dbr.as_ptr().add(0x2c) as *const u32),
                fsinfo_sector: *(dbr.as_ptr().add(0x30) as *const u16),
            }
        }
    }

    pub fn bytes_per_cluster(&self) -> u32 {
        self.bytes_per_sector as u32 * self.sectors_per_cluster as u32
    }
}

// 该结构体不对Buffer作结构映射，仅保留位置信息
// 但是为其中信息的获取和修改提供了接口
pub struct FSInfo {
    sector_num: u32,
}

impl FSInfo {
    pub fn new(sector_num: u32) -> Self {
        Self { sector_num }
    }

    fn check_lead_signature(&self, block_device: Arc<dyn BlockDevice>) -> bool {
        get_block_cache(self.sector_num as usize, block_device)
            .lock()
            .read(0, |&lead_sig: &u32| lead_sig == LEAD_SIGNATURE)
    }

    fn check_another_signature(&self, block_device: Arc<dyn BlockDevice>) -> bool {
        get_block_cache(self.sector_num as usize, block_device)
            .lock()
            .read(484, |&sec_sig: &u32| sec_sig == SECOND_SIGNATURE)
    }

    /*对签名进行校验*/
    pub fn check_signature(&self, block_device: Arc<dyn BlockDevice>) -> bool {
        return self.check_lead_signature(block_device.clone())
            && self.check_another_signature(block_device.clone());
    }

    /*读取空闲簇数*/
    pub fn read_free_clusters(&self, block_device: Arc<dyn BlockDevice>) -> u32 {
        get_block_cache(self.sector_num as usize, block_device)
            .lock()
            .read(488, |&free_cluster_count: &u32| free_cluster_count)
    }

    /*写空闲簇数*/
    pub fn write_free_clusters(&self, free_clusters: u32, block_device: Arc<dyn BlockDevice>) {
        get_block_cache(self.sector_num as usize, block_device)
            .lock()
            .modify(488, |free_cluster_count: &mut u32| {
                *free_cluster_count = free_clusters;
            });
    }

    // 下一个可用簇号，通常设置为最后一个分配的簇号
    pub fn read_first_free_cluster(&self, block_device: Arc<dyn BlockDevice>) -> u32 {
        get_block_cache(self.sector_num as usize, block_device)
            .lock()
            .read(492, |&start_cluster: &u32| start_cluster)
    }

    // 更新下一个可用簇号
    pub fn write_first_free_cluster(&self, start_cluster: u32, block_device: Arc<dyn BlockDevice>) {
        get_block_cache(self.sector_num as usize, block_device)
            .lock()
            .modify(492, |start_clu: &mut u32| {
                *start_clu = start_cluster;
            });
    }
}

// 不使用第二个 FAT 表
#[derive(Clone, Copy)]
pub struct FAT {
    fat1_sector: u32, // FAT1起始扇区
    sectors: u32,     // 一个 FAT 表的大小
    entries: u32,     // 每扇区有多少个 FAT 表项
}

impl FAT {
    pub fn new(fat1_sector: u32, sectors: u32, entries: u32) -> Self {
        Self {
            fat1_sector,
            sectors,
            entries,
        }
    }

    /* 计算簇对应表项的位置：sector 和 offset */
    fn calculate_pos(&self, cluster: u32) -> (u32, u32) {
        let fat1_sec = self.fat1_sector + cluster / FATENTRY_PER_SEC;
        let offset = 4 * (cluster % FATENTRY_PER_SEC);
        (fat1_sec, offset)
    }

    // 搜索下一个可用簇
    // 找不到下一个可用簇则返回 0
    pub fn next_free_cluster(
        &self,
        current_cluster: u32,
        block_device: Arc<dyn BlockDevice>,
    ) -> u32 {
        let total_cluster = self.sectors * self.entries;
        let mut current_cluster = current_cluster + 1;
        for i in 0..total_cluster {
            current_cluster = (current_cluster + i) % total_cluster;
            if current_cluster == 0 || current_cluster == 1 {
                continue;
            }

            let (fat1_sec, offset) = self.calculate_pos(current_cluster);
            // 查看当前cluster的表项
            let entry_val = get_block_cache(fat1_sec as usize, block_device.clone())
                .lock()
                .read(offset as usize, |&entry_val: &u32| entry_val);
            if entry_val == FREE_CLUSTER {
                return current_cluster & 0x0FFFFFFF;
            }
        }
        0
    }

    /// 查询当前簇的下一个簇
    ///
    /// 失败返回 0
    pub fn get_next_cluster(&self, cluster: u32, block_device: Arc<dyn BlockDevice>) -> u32 {
        let (fat1_sec, offset) = self.calculate_pos(cluster);
        let fat1_rs = get_block_cache(fat1_sec as usize, block_device.clone())
            .lock()
            .read(offset as usize, |&next_cluster: &u32| next_cluster);
        if fat1_rs == BAD_CLUSTER || fat1_rs >= BOUND_CLUSTER {
            0
        } else {
            fat1_rs & 0x0FFFFFFF
        }
    }

    /* 设置当前簇的下一个簇 */
    pub fn set_next_cluster(
        &self,
        cluster: u32,
        next_cluster: u32,
        block_device: Arc<dyn BlockDevice>,
    ) {
        let (fat1_sec, offset) = self.calculate_pos(cluster);
        get_block_cache(fat1_sec as usize, block_device.clone())
            .lock()
            .modify(offset as usize, |old_clu: &mut u32| {
                *old_clu = next_cluster;
            });
    }

    pub fn set_end(&self, cluster: u32, block_device: Arc<dyn BlockDevice>) {
        self.set_next_cluster(cluster, END_CLUSTER, block_device);
    }

    // 从 start_cluster 开始的第 index 个簇号
    // index = 1 表示下一个簇号
    // 失败返回 0
    pub fn get_cluster_at(
        &self,
        start_cluster: u32,
        index: u32,
        block_device: Arc<dyn BlockDevice>,
    ) -> u32 {
        let mut cluster = start_cluster;
        for _ in 0..index {
            cluster = self.get_next_cluster(cluster, block_device.clone());
            if cluster == 0 {
                break;
            }
        }
        cluster & 0x0FFFFFFF
    }

    /// 获得从 start_cluster（含）开始的所有簇链</br>
    /// 假定 start_cluster 始终有效
    pub fn get_all_clusters(
        &self,
        start_cluster: u32,
        block_device: Arc<dyn BlockDevice>,
    ) -> Vec<u32> {
        let mut curr_cluster = start_cluster;
        let mut v_cluster: Vec<u32> = Vec::new();
        if start_cluster >= BOUND_CLUSTER {
            return v_cluster;
        }
        loop {
            v_cluster.push(curr_cluster & 0x0FFFFFFF);
            let next_cluster = self.get_next_cluster(curr_cluster, block_device.clone());
            if next_cluster == 0 {
                return v_cluster;
            } else {
                curr_cluster = next_cluster;
            }
        }
    }

    pub fn get_final_cluster(&self, start_cluster: u32, block_device: Arc<dyn BlockDevice>) -> u32 {
        // 同样没有判断 start_cluster 的有效值
        self.get_all_clusters(start_cluster, block_device)
            .pop()
            .unwrap()
    }

    pub fn count_claster_num(&self, start_cluster: u32, block_device: Arc<dyn BlockDevice>) -> u32 {
        if start_cluster == 0 {
            return 0;
        }
        self.get_all_clusters(start_cluster, block_device).len() as u32
    }
}
