// use crate::{BLOCK_SZ, BOUND_CLUSTER, FAT};
use crate::{println, BLOCK_SZ, BOUND_CLUSTER, END_CLUSTER, FAT};

use super::{dir_entry::*, fat32_manager::*, get_block_cache, BlockDevice};
use alloc::borrow::ToOwned;
use alloc::string::String;
use alloc::sync::Arc;
use alloc::vec::Vec;
use spin::RwLock;

type DataBlock = [u8; BLOCK_SZ];

/// 文件 inode 结点
///
/// 对于 fat32 文件系统，一个文件可以由簇号识别，对应文件的数据区。
/// 对于普通文件来说，可能需要更改文件大小、时间等信息，而这存储在目录项中。
/// 因此 inode 存储文件对应的目录项，同时保存父目录的首簇号以及该目录项在父目录的偏移，
/// 从而确定该目录项的位置，这样设计也是为了便于获得文件名以及文件路径。
///
/// 对于目录等文件来说，其可能对应几个目录项（目录文件本身目录项以及`.`, `..`目录项），
/// 因此不能通过扇区号和偏移量来区分文件，只能通过首簇号区分文件，而除了目录文件，
/// 普通文件不会预先分配空间，此时得到的首簇号都是 END_CLUSTER，这时将通过绝对路径区分文件
///
/// 对于根目录，其`dir_first_cluster = 0`，`dir_offset`没有意义
///
/// 对于目录项 short_dir_entry 的处理，由于所需更改的字段较少（目前只有 size 这一字段）且所占空间不大，
/// 将其复制到内存，并设置 dirt 位，只在必要的时候进行更新。
///
/// 虽然成员 fs 包含 fat 以及 block_device，但将其分离出来方便处理
#[derive(Clone)]
pub struct Inode {
    dir_first_cluster: u32,         // 文件父目录数据的首簇号
    dir_offset: usize,              // 文件短目录项在文件父目录数据区的偏移
    short_dir_entry: ShortDirEntry, // 文件短目录项
    dirt: bool,                     // 短目录项的脏位（判断短目录项是否有改变，而不是文件数据）
    fs: Arc<RwLock<FAT32Manager>>,
    fat: Arc<RwLock<FAT>>,
    block_device: Arc<dyn BlockDevice>,
}

impl Inode {
    /// 新建 inode 结点
    pub fn new(
        dir_first_cluster: u32,
        dir_offset: usize,
        short_dir_entry: ShortDirEntry,
        fs: Arc<RwLock<FAT32Manager>>,
        fat: Arc<RwLock<FAT>>,
        block_device: Arc<dyn BlockDevice>,
    ) -> Self {
        Self {
            dir_first_cluster,
            dir_offset,
            short_dir_entry,
            dirt: false,
            fs,
            fat,
            block_device,
        }
    }

    /// 判断文件是否为目录文件
    pub fn is_dir(&self) -> bool {
        self.short_dir_entry.is_dir()
    }

    /// 获取文件属性
    pub fn attribute(&self) -> Attribute {
        Attribute::from_bits(self.short_dir_entry.attribute).unwrap()
    }

    /// 读取可能的长文件目录项，返回长文件名以及第一个长文件目录项的偏移
    ///
    /// 由于存储方式可能不一样，不能直接根据文件名是长短文件名来判断
    fn get_long_name(&self, checksum: u8) -> (Option<String>, usize) {
        let fs = self.fs.read();
        let mut num = 1;
        let mut offset = self.dir_offset as isize;
        let mut entry = LongDirEntry::empty();
        let mut long_name = String::new();
        loop {
            offset -= DIRENT_SZ as isize;
            if offset < 0 {
                break;
            }
            let (_, sector, block_offset) = fs.get_pos(self.dir_first_cluster, offset as usize);
            get_block_cache(sector, Arc::clone(&self.block_device))
                .lock()
                .read(block_offset, |tmp: &LongDirEntry| {
                    entry = *tmp;
                });
            if !entry.is_valid() || entry.checksum != checksum || (entry.order & (!0x40)) != num {
                break;
            }
            long_name += &entry.get_name_format();
            if (entry.order & 0x40) != 0 {
                return (Some(long_name), offset as usize);
            }
            num += 1;
        }
        (None, 0)
    }

    /// 获取文件名
    ///
    /// 有的目录项 (`.` `..`) 并不对应真正的文件名，需要读取文件真正位置的目录项
    pub fn name(&self) -> String {
        let fs = self.fs.read();

        // 读取真正的短文件目录项
        // 只有目录文件才可能出现
        let mut short_entry = self.short_dir_entry.clone();
        if self.attribute().contains(Attribute::DIRECTORY) {
            // 父目录首簇为0，说明当前文件是根文件
            if self.dir_first_cluster == 0 {
                return String::from("/");
            }

            let (_, sector, block_offset) = fs.get_pos(self.dir_first_cluster, self.dir_offset);
            get_block_cache(sector, Arc::clone(&self.block_device))
                .lock()
                .read(block_offset, |tmp: &ShortDirEntry| {
                    short_entry = *tmp;
                });
        }

        if let (Some(long_name), _) = self.get_long_name(short_entry.checksum()) {
            long_name
        } else {
            short_entry.name()
        }
    }

    /// 获取文件大小
    pub fn size(&self) -> usize {
        if self.is_dir() {
            (self
                .fat
                .read()
                .count_claster_num(self.first_cluster(), Arc::clone(&self.block_device))
                * self.fs.read().bytes_per_cluster) as usize
        } else {
            self.short_dir_entry.size as usize
        }
    }

    /// 设置大小
    pub fn set_size(&mut self, size: usize) {
        if !self.is_dir() {
            self.short_dir_entry.set_size(size as u32);
            self.dirt = true;
        }
    }

    /// 获取文件数据区第一个簇号
    pub fn first_cluster(&self) -> u32 {
        self.short_dir_entry.first_cluster()
    }

    /// 设置首簇
    pub fn set_first_cluster(&mut self, cluster: u32) {
        self.short_dir_entry.set_first_cluster(cluster);
        self.dirt = true;
    }

    /// 扩容
    fn increase_size(&mut self, new_size: usize) {
        let fs = self.fs.read();
        let num = fs.cluster_num_needed(self.size(), new_size);
        if num != 0 {
            match fs.alloc_cluster(num as u32) {
                0 => {
                    panic!("no space!");
                }
                first_cluster => {
                    let old_first_cluster = self.first_cluster();
                    if old_first_cluster >= BOUND_CLUSTER {
                        drop(fs);
                        self.set_first_cluster(first_cluster);
                        self.set_size(new_size);
                        return;
                    } else {
                        let fat = self.fat.write();
                        let old_final_cluster = fat
                            .get_final_cluster(old_first_cluster, Arc::clone(&self.block_device));
                        fat.set_next_cluster(
                            old_final_cluster,
                            first_cluster,
                            Arc::clone(&self.block_device),
                        );
                    }
                }
            }
        }
        drop(fs);
        self.set_size(new_size);
    }

    /// 分配连续的目录项空间，用来保证长短文件目录项相邻
    ///
    /// num: 目录项的个数</br>
    /// 返回分配的第一个目录项的偏移</br>
    /// 简化：直接从目录的最后位置进行查找扩容，而不从头查找是否有文件被删除从而空出空间
    fn alloc_entry(&mut self, num: usize) -> usize {
        let bytes_per_cluster = self.fs.read().bytes_per_cluster as usize;

        // 获得目录最后一个簇的偏移
        // 由于正常情况下只由目录文件调用，其大小肯定为簇大小的整数倍，
        // 因此减掉一个簇大小即为最后一个簇的偏移
        let offset = self.size() - bytes_per_cluster;

        let mut buf = vec![0u8; bytes_per_cluster];
        let buf = buf.as_mut_slice();

        // 查找最后空闲的目录项
        self.read_at(offset, buf);
        let mut index = 0;
        while index < bytes_per_cluster {
            if buf[index] == 0 {
                break;
            }
            index += DIRENT_SZ;
        }
        if index + num * DIRENT_SZ > bytes_per_cluster {
            self.increase_size(offset + index + num * DIRENT_SZ);
        }
        offset + index
    }

    /// 在当前目录下创建指定文件，若该文件存在则返回 None，否则创建文件
    pub fn create(&mut self, name: &str, attribute: Attribute) -> Option<Inode> {
        // println!("in create");
        if !self.is_dir() || self.find_entry(name).is_some() {
            return None;
        }
        let fs = self.fs.read();
        let mut entries = Vec::new();

        // 为目录文件分配首簇
        let cluster = if attribute.contains(Attribute::DIRECTORY) {
            fs.alloc_cluster(1)
        } else {
            END_CLUSTER
        };

        // 构造目录项数组
        let entry;
        if fs.is_short_name(name) {
            let (primary, ext) = fs.short_name_format(name);
            entry = ShortDirEntry::new(primary, ext, attribute.bits(), cluster);
            entries.push(entry.clone().as_bytes().to_owned());
        } else {
            let short_name = fs.generate_short_name(name);
            // println!("generate shor name = {}", short_name);
            let (primary, ext) = fs.short_name_format(&short_name);
            entry = ShortDirEntry::new(primary, ext, attribute.bits(), cluster);
            let checksum = entry.checksum();
            let name_split = fs.long_name_format(name);
            for i in (1..=name_split.len()).rev() {
                if i == name_split.len() {
                    let long = LongDirEntry::new(&name_split[i - 1], i as u8 | 0x40, checksum);
                    entries.push(long.as_bytes().to_owned());
                } else {
                    let long = LongDirEntry::new(&name_split[i - 1], i as u8, checksum);
                    entries.push(long.as_bytes().to_owned());
                }
            }
            entries.push(entry.clone().as_bytes().to_owned());
        }

        // 查找可以插入的位置，保证长短文件目录项相邻
        // 可能这里需要再次加锁
        drop(fs);
        let offset = self.alloc_entry(entries.len());
        let buf = entries.concat();
        self.write_at(offset, buf.as_slice());

        let mut res = Inode::new(
            self.first_cluster(),
            offset + (entries.len() - 1) * DIRENT_SZ,
            entry,
            Arc::clone(&self.fs),
            Arc::clone(&self.fat),
            Arc::clone(&self.block_device),
        );

        // 如果创建的文件是目录，则先写入 . 和 .. 目录项
        if attribute.contains(Attribute::DIRECTORY) {
            let fs = self.fs.read();
            let (primary, ext) = fs.short_name_format(".");
            let dot = ShortDirEntry::new(primary, ext, attribute.bits(), cluster);
            let (primary, ext) = fs.short_name_format("..");
            let dotdot =
                ShortDirEntry::new(primary, ext, self.attribute().bits(), self.first_cluster());
            let mut buf: Vec<u8> = Vec::new();
            buf.extend_from_slice(dot.as_bytes());
            buf.extend_from_slice(dotdot.as_bytes());
            res.write_at(0, buf.as_slice());
        }

        // println!(
        //     "create file: name = {}, dir_first_cluster = {}, file_first_cluster = {}",
        //     res.name(),
        //     res.dir_first_cluster,
        //     res.first_cluster()
        // );
        Some(res)
    }

    /// 读文件
    pub fn read_at(&self, offset: usize, buf: &mut [u8]) -> usize {
        let fs = self.fs.read();
        let fat = self.fat.read();
        let bytes_per_sector = fs.bytes_per_sector as usize;
        let bytes_per_cluster = fs.bytes_per_cluster as usize;
        let mut current_off = offset;
        let end: usize;

        if self.is_dir() {
            let size = bytes_per_cluster
                * fat.count_claster_num(self.first_cluster() as u32, Arc::clone(&self.block_device))
                    as usize;
            end = offset + buf.len().min(size);
        } else {
            end = (offset + buf.len()).min(self.size() as usize);
        }
        if current_off >= end {
            return 0;
        }

        let (mut current_cluster, mut current_sector, _) = fs.get_pos(self.first_cluster(), offset);
        if current_cluster == 0 {
            return 0;
        };

        let mut read_size = 0usize;
        loop {
            // 将偏移量向上对齐扇区大小
            let mut end_current_block = (current_off / bytes_per_sector + 1) * bytes_per_sector;
            end_current_block = end_current_block.min(end);

            // 读
            let block_read_size = end_current_block - current_off;
            let dst = &mut buf[read_size..read_size + block_read_size];
            get_block_cache(current_sector, Arc::clone(&self.block_device))
                .lock()
                .read(0, |data_block: &DataBlock| {
                    dst.copy_from_slice(
                        &data_block
                            [current_off % BLOCK_SZ..current_off % BLOCK_SZ + block_read_size],
                    );
                });

            // 更新读取长度
            read_size += block_read_size;
            if end_current_block == end {
                break;
            }

            // 更新索引参数
            current_off = end_current_block;
            if current_off % bytes_per_cluster == 0 {
                // 读完一个簇
                current_cluster =
                    fat.get_next_cluster(current_cluster, Arc::clone(&self.block_device));
                if current_cluster == 0 {
                    break;
                }
                // 计算所在扇区
                current_sector = fs.first_sector_of_cluster(current_cluster);
            } else {
                current_sector += 1; //没读完一个簇，直接进入下一扇区
            }
        }
        read_size
    }

    /// 写文件
    pub fn write_at(&mut self, offset: usize, buf: &[u8]) -> usize {
        if self.attribute().contains(Attribute::RDONLY) {
            println!("can't write the file: read only!");
            return 0;
        }
        self.increase_size(offset + buf.len());
        // 获取共享锁
        let fs = self.fs.read();
        let fat = self.fat.read();
        let bytes_per_sector = fs.bytes_per_sector as usize;
        let bytes_per_cluster = fs.bytes_per_cluster as usize;
        let mut current_off = offset;
        let end: usize;

        if self.is_dir() {
            let size = bytes_per_cluster
                * fat.count_claster_num(self.first_cluster() as u32, Arc::clone(&self.block_device))
                    as usize;
            end = offset + buf.len().min(size); // DEBUG:约束上界
        } else {
            end = (offset + buf.len()).min(self.size() as usize);
        }

        let (mut current_cluster, mut current_sector, _) = fs.get_pos(self.first_cluster(), offset);
        let mut write_size = 0usize;

        loop {
            // 将偏移量向上对齐扇区大小（一般是512
            let mut end_current_block = (current_off / bytes_per_sector + 1) * bytes_per_sector;
            end_current_block = end_current_block.min(end);

            // 写
            let block_write_size = end_current_block - current_off;
            get_block_cache(current_sector, Arc::clone(&self.block_device))
                .lock()
                .modify(0, |data_block: &mut DataBlock| {
                    let dst = &mut data_block
                        [current_off % BLOCK_SZ..current_off % BLOCK_SZ + block_write_size];
                    dst.copy_from_slice(&buf[write_size..write_size + block_write_size]);
                });

            // 更新读取长度
            write_size += block_write_size;
            if end_current_block == end {
                break;
            }

            // 更新索引参数
            current_off = end_current_block;
            if current_off % bytes_per_cluster == 0 {
                // 读完一个簇
                current_cluster =
                    fat.get_next_cluster(current_cluster, Arc::clone(&self.block_device));
                if current_cluster == 0 {
                    panic!("END_CLUSTER");
                }
                // 计算所在扇区
                current_sector = fs.first_sector_of_cluster(current_cluster);
            } else {
                current_sector += 1; //没读完一个簇，直接进入下一扇区
            }
        }
        write_size
    }

    /// 清空文件
    ///
    /// 这里没有对长文件目录项清空，应该不需要吧
    pub fn clear(&mut self) {
        let fat = self.fat.write();
        let clusters = fat.get_all_clusters(self.first_cluster(), Arc::clone(&self.block_device));
        drop(fat);
        let fs = self.fs.read();
        fs.dealloc_cluster(clusters);
        drop(fs);
        self.short_dir_entry.clear();
        self.dirt = true;
    }

    /// 删除文件
    pub fn remove(&mut self) {
        let fs = self.fs.read();
        if let (Some(_), mut offset) = self.get_long_name(self.short_dir_entry.checksum()) {
            while offset < self.dir_offset {
                // println!("remove long dirent: offset = {:#x}", offset);
                let (_, sector, block_offset) = fs.get_pos(self.dir_first_cluster, offset);
                get_block_cache(sector, Arc::clone(&self.block_device))
                    .lock()
                    .modify(block_offset, |tmp: &mut LongDirEntry| {
                        tmp.delete();
                    });
                offset += DIRENT_SZ;
            }
        }
        drop(fs);
        self.clear();
        // println!("remove short dirent");
        self.short_dir_entry.delete();
    }
}

/// 该 impl 块实现相关 inode 结点的查找
/// 在查找的时候注意对于短文件目录项，文件名不区分大小写，都是大写的，
/// 对于长文件目录项，则区分大小写。
/// 因此不能轻易对路径转大写
impl Inode {
    /// 解析当前目录偏移 offset 下的长文件目录项
    /// 需保证长文件目录项与对应的短文件目录项相邻
    ///
    /// offset: 长文件目录项的第一个目录项的偏移
    ///
    /// 返回值：</br>
    /// 成功: (长文件目录项对应的短文件目录项，短文件目录项下一项的偏移，长文件名)</br>
    /// 失败: (None，失败的目录项偏移，无意义字符串)
    fn parse_long_dir_entry(&self, offset: usize) -> (Option<Inode>, usize, String) {
        let mut name = String::new();
        let mut tmp = LongDirEntry::empty();

        // 解析长文件目录项，校验并获得长文件名
        self.read_at(offset, tmp.as_bytes_mut());
        if !tmp.is_valid() || (tmp.order & 0x40 == 0) {
            return (None, offset, name);
        }
        name = tmp.get_name_format() + &name;
        let order = (tmp.order & (!0x40)) as usize;
        let checksum = tmp.checksum;
        for i in 1..order {
            self.read_at(offset + i * DIRENT_SZ, tmp.as_bytes_mut());
            if !tmp.is_valid() || tmp.checksum != checksum {
                return (None, offset + i * DIRENT_SZ, name);
            }
            name = tmp.get_name_format() + &name;
        }

        // 获取对应的短文件名，并做最终校验
        let mut tmp = ShortDirEntry::empty();
        self.read_at(offset + order * DIRENT_SZ, tmp.as_bytes_mut());
        // println!("parse long name: get short name = {}", tmp.name());
        // println!("parse long name: bytes = {:?}", tmp.as_bytes());
        if !tmp.is_valid() || tmp.checksum() != checksum {
            return (None, offset + order * DIRENT_SZ, name);
        }
        // println!("parse long name: {}", name);
        (
            Some(Inode::new(
                self.first_cluster(),
                offset + order * DIRENT_SZ,
                tmp,
                Arc::clone(&self.fs),
                Arc::clone(&self.fat),
                Arc::clone(&self.block_device),
            )),
            offset + (order + 1) * DIRENT_SZ,
            name,
        )
    }

    /// 根据首簇号确定文件，获取文件在该数据区的偏移
    ///
    /// 首簇号决定了某个文件的数据区</br>
    /// 由于每个创建的文件都在初始化时分配簇，因此能够通过首簇号区分文件
    fn get_offset(&self, cluster: u32) -> usize {
        let mut offset = 0;
        let mut entry = ShortDirEntry::empty();
        while offset < self.size() {
            self.read_at(offset, entry.as_bytes_mut());
            if entry.first_cluster() == cluster {
                break;
            }
            offset += DIRENT_SZ;
        }
        offset
    }

    /// 在当前目录中查找指定文件
    ///
    /// 对于普通文件，允许`.` `..`路径的查找，也方便获得完整路径重用代码
    ///
    /// 注：在查找的时候不能直接根据文件名判断它是以长文件还是短文件形式存储，然后分类查找。
    /// 因为有的系统即使为短文件也以长文件形式存储（如 linux 挂载 fat32 文件系统），这时根据
    /// 长文件形成的短文件名不一定与本文件名相同。同样也不能提前根据文件名来对文件名进行相应的格式转换。
    ///
    /// # example
    ///
    /// 文件名`abcdefgh`为短文件名，但假若之前已经创建了一个长文件名`abcdefghijk`，则会先
    /// 自动生成一个短文件名（生成算法不一定，但本文件系统的算法如此）`ABCDEFGH`。因此以长文件形式存储时，
    /// 再创建所需文件名`abcdefgh`，则会自动生成短文件名`ABCDEF~1`，此时根据短文件名查找失败。
    pub fn find_entry(&self, name: &str) -> Option<Inode> {
        // println!("in find_entry: name = {:?}", name);
        let mut tmp = ShortDirEntry::empty();
        match name {
            "." => return Some(self.clone()),
            ".." => {
                if self.first_cluster() == 2 {
                    return Some(self.clone());
                }

                // 如果父目录是根节点，直接返回
                let dir_first_cluster = self.dir_first_cluster;
                if dir_first_cluster == 0 || dir_first_cluster == 2 {
                    return Some(FAT32Manager::root_inode(&self.fs));
                }

                // 读取父目录
                let (_, sector, block_offset) = self.fs.read().get_pos(dir_first_cluster, 0);
                get_block_cache(sector, Arc::clone(&self.block_device))
                    .lock()
                    .read(block_offset, |entry: &ShortDirEntry| {
                        tmp = *entry;
                    });
                let mut res = Inode::new(
                    0,
                    0,
                    tmp,
                    Arc::clone(&self.fs),
                    Arc::clone(&self.fat),
                    Arc::clone(&self.block_device),
                );

                // 读取父目录的父目录
                let mut ppentry = ShortDirEntry::empty();
                res.read_at(0x20, ppentry.as_bytes_mut());
                res.dir_first_cluster = ppentry.first_cluster();
                res.dir_offset = Inode::new(
                    0,
                    0,
                    ppentry,
                    Arc::clone(&self.fs),
                    Arc::clone(&self.fat),
                    Arc::clone(&self.block_device),
                )
                .get_offset(res.first_cluster());

                return Some(res);
            }
            _ => {
                let mut offset = 0;
                let size = self.size();
                while offset < size {
                    self.read_at(offset, tmp.as_bytes_mut());
                    if tmp.is_free() {
                        break;
                    }
                    if tmp.is_valid() {
                        if tmp.is_short() {
                            if tmp.name() == name.to_ascii_uppercase() {
                                return Some(Inode::new(
                                    self.first_cluster(),
                                    offset,
                                    tmp,
                                    Arc::clone(&self.fs),
                                    Arc::clone(&self.fat),
                                    Arc::clone(&self.block_device),
                                ));
                            }
                            offset += DIRENT_SZ;
                        } else {
                            let (res, off, long_name) = self.parse_long_dir_entry(offset);
                            if let Some(res) = res {
                                // println!("get long name = {:?}", long_name);
                                // println!("find long name = {:?}", name);
                                if long_name == name {
                                    return Some(res);
                                }
                            }
                            offset = off;
                        }
                    } else {
                        offset += DIRENT_SZ;
                    }
                }
            }
        }
        // println!("out find_entry: name = {}", name);
        None
    }

    /// 以当前目录为起点，根据路径进行搜索，直到达到最顶层目录
    ///
    /// 返回目录 inode 结点
    pub fn get_dir(&self, path: &str) -> Option<Inode> {
        let mut parent = self.clone();
        let mut path = path.trim_end_matches('/').split('/').collect::<Vec<_>>();
        path.pop();
        for name in path {
            // println!("in get_dir: name = {}", name);
            if name.is_empty() {
                continue;
            }
            match parent.find_entry(name) {
                Some(res) => parent = res,
                None => return None,
            }
        }
        Some(parent)
    }

    /// 以当前目录为起点，根据路径查找 inode 结点
    pub fn namei(&self, path: &str) -> Option<Inode> {
        if let Some(name) = path
            .trim_end_matches('/')
            .split('/')
            .collect::<Vec<_>>()
            .pop()
        {
            match self.get_dir(path) {
                Some(parent) => parent.find_entry(name),
                None => None,
            }
        } else {
            None
        }
    }

    /// 获取当前目录下从偏移量 offset 开始的所有文件信息
    ///
    /// 返回值：(短文件目录项，文件名，下一个目录项的偏移)
    ///
    /// todo: 根据显示规则对短文件名进行转换，现在默认全部大写
    pub fn list(&self, mut offset: usize) -> Vec<(ShortDirEntry, String, usize)> {
        let mut res = Vec::new();
        let size = self.size();
        let mut tmp = ShortDirEntry::empty();
        while offset < size {
            self.read_at(offset, tmp.as_bytes_mut());
            if tmp.is_free() {
                break;
            }
            if tmp.is_valid() {
                if tmp.is_short() {
                    offset += DIRENT_SZ;
                    res.push((tmp.clone(), tmp.name(), offset));
                } else {
                    let (dirent, off, long_name) = self.parse_long_dir_entry(offset);
                    if dirent.is_some() {
                        // println!("ls: get long name = {}, next offset = {:#x}", long_name, off);
                        res.push((dirent.unwrap().short_dir_entry.clone(), long_name, off));
                    }
                    offset = off;
                }
            } else {
                offset += DIRENT_SZ;
            }
        }
        res
    }

    /// 获取从根目录到该文件的路径
    pub fn path(&self) -> String {
        let mut res = self.name();
        if res == "/" {
            return res;
        }
        let mut tmp = res.clone();
        let mut inode = self.clone();
        loop {
            inode = inode.get_dir(&(String::from("../") + &tmp)).unwrap();
            res = String::from("/") + &res;
            if inode.first_cluster() == 2 {
                break;
            }
            tmp = inode.name();
            res = tmp.clone() + &res;
        }
        res
    }
}

impl Drop for Inode {
    fn drop(&mut self) {
        // println!("drop inode, name = {}", self.name());
        if self.dir_first_cluster != 0 && self.dirt {
            // println!("drop inode: write to disk");
            let (_, sector, block_offset) = self
                .fs
                .read()
                .get_pos(self.dir_first_cluster, self.dir_offset);
            get_block_cache(sector, Arc::clone(&self.block_device))
                .lock()
                .modify(block_offset, |entry: &mut ShortDirEntry| {
                    *entry = self.short_dir_entry
                });
        }
    }
}
