use alloc::string::String;
use fat32::BLOCK_SZ;

const MAXNAME: usize = 128;
pub const DT_DIR: u8 = 4;
pub const DT_REG: u8 = 8;

#[repr(C)]
#[derive(Debug)]
pub struct Dirent {
    pub d_ino: u64, // 索引结点号(第一个簇号)
    pub d_off: i64, // 64-bit offset to next structure

    // Size of this dirent，在 man 文档中，dirent 结构体中 d_name 大小是不确定的，没有规定 MAXNAME 的大小
    // 通过设置 null 来表示结尾，因此可能用该变量表示该结构体的大小
    pub d_reclen: u16,

    pub d_type: u8,            // 文件类型
    pub d_name: [u8; MAXNAME], // 文件名
}

impl Dirent {
    pub fn new(mut name: String, firstcluster: u64, attribute: u8, offset: usize) -> Self {
        // 由于 k210 的对齐要求，需要进行 64 位的对齐
        let mut len = name.len() + 19;
        let align = 8 - len % 8;
        len += align;
        for _ in 0..align {
            name.push('\0');
        }
        Self {
            d_ino: firstcluster,
            d_off: offset as i64,
            d_reclen: len as u16,
            d_type: attribute,
            d_name: {
                let mut tmp: [u8; MAXNAME] = [0; MAXNAME];
                tmp[..name.len()].copy_from_slice(name.as_bytes());
                tmp
            },
        }
    }

    pub fn as_bytes(&self) -> &[u8] {
        let size = self.d_reclen as usize;
        // println!("in dirent: size = {}", size);
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, size) }
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct Kstat {
    pub st_dev: u64,
    pub st_ino: u64,
    pub st_mode: u32,
    pub st_nlink: u32,
    pub st_uid: u32,
    pub st_gid: u32,
    pub st_rdev: u64,
    __pad: u64,
    pub st_size: i64,
    pub st_blksize: u32,
    __pad2: i32,
    pub st_blocks: u64,
    pub st_atime_sec: i64,
    pub st_atime_nsec: i64,
    pub st_mtime_sec: i64,
    pub st_mtime_nsec: i64,
    pub st_ctime_sec: i64,
    pub st_ctime_nsec: i64,
    __unused: [u32; 2],
}

impl Kstat {
    pub fn new(st_dev: u64, st_ino: u64, st_mode: u32, st_size: i64) -> Self {
        Self {
            st_dev,
            st_ino,
            st_mode,
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size,
            st_blksize: BLOCK_SZ as u32,
            __pad2: 0,
            st_blocks: ((st_size as usize + BLOCK_SZ - 1) / BLOCK_SZ) as u64,
            st_atime_sec: 0,
            st_atime_nsec: 0,
            st_mtime_sec: 0,
            st_mtime_nsec: 0,
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }

    pub fn as_bytes(&self) -> &[u8] {
        let size = core::mem::size_of::<Self>();
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, size) }
    }
}

#[repr(C)]
pub struct IOVec {
    pub iov_base: u64,
    pub iov_len: usize,
}

#[repr(C)]
pub struct Winsize {
    ws_row: u16,
    ws_col: u16,
    ws_xpixel: u16,
    ws_ypixel: u16,
}

impl Winsize {
    pub fn new() -> Self {
        Self {
            ws_row: 24,
            ws_col: 80,
            ws_xpixel: 0,
            ws_ypixel: 0,
        }
    }
}
