use super::info::{Dirent, DT_DIR, DT_REG};
use super::{remove_from_file_table, File, FileClass, Kstat, ROOT_INODE};
use crate::fs::insert_file_table;
use crate::mm::UserBuffer;
use crate::task::current_task;
use alloc::sync::Arc;
use alloc::vec;
use alloc::vec::Vec;
use bitflags::*;
use fat32::{Attribute, Inode};
use spin::{Mutex, MutexGuard, RwLock};

pub const AT_FDCWD: isize = -100;

pub struct OSInode {
    readable: bool,
    writable: bool,
    inner: Mutex<OSInodeInner>,
}

pub struct OSInodeInner {
    offset: usize,
    pub inode: Arc<RwLock<Inode>>,
}

impl OSInode {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, OSInodeInner> {
        self.inner.lock()
    }

    pub fn new(readable: bool, writable: bool, inode: Arc<RwLock<Inode>>) -> Self {
        Self {
            readable,
            writable,
            inner: Mutex::new(OSInodeInner { offset: 0, inode }),
        }
    }
    pub fn read_all(&self) -> Vec<u8> {
        let mut inner = self.inner_exclusive_access();
        let mut buffer = [0u8; 512];
        let mut v: Vec<u8> = Vec::new();
        loop {
            let len = inner.inode.read().read_at(inner.offset, &mut buffer);
            if len == 0 {
                break;
            }
            inner.offset += len;
            v.extend_from_slice(&buffer[..len]);
        }
        v
    }

    pub fn offset(&self) -> usize {
        self.inner_exclusive_access().offset
    }

    pub fn seek(&self, off: usize) {
        self.inner_exclusive_access().offset = off;
    }

    pub fn size(&self) -> usize {
        self.inner_exclusive_access().inode.read().size()
    }

    pub fn getdents64(&self) -> Vec<u8> {
        let mut res = Vec::new();
        let mut inner = self.inner_exclusive_access();
        let mut end = inner.offset;
        let inode = inner.inode.read();
        for (dirent, mut name, offset) in inode.list(end) {
            name.push('\0');
            let attribute = if dirent.is_dir() { DT_DIR } else { DT_REG };
            let tmp = Dirent::new(name, dirent.first_cluster() as u64, attribute, offset);
            res.extend_from_slice(tmp.as_bytes());
            end = offset;
        }
        drop(inode);
        inner.offset = end;
        res
    }

    pub fn fstat(&self) -> Kstat {
        let inner = self.inner_exclusive_access();
        let inode = inner.inode.read();
        let size = inode.size();
        let mut mode = 0o0777;
        if inode.is_dir() {
            mode |= 0o40000;
        } else {
            mode |= 0o100000;
        }
        Kstat::new(0, inode.first_cluster() as u64, mode, size as i64)
    }

    /// 清空文件
    pub fn clear(&self) {
        self.inner_exclusive_access().inode.write().clear();
    }
}

impl File for OSInode {
    fn readable(&self) -> bool {
        self.readable
    }
    fn writable(&self) -> bool {
        self.writable
    }
    fn read(&self, mut buf: UserBuffer) -> usize {
        let mut inner = self.inner_exclusive_access();
        let mut total_read_size = 0usize;
        for slice in buf.buffers.iter_mut() {
            let read_size = inner.inode.read().read_at(inner.offset, *slice);
            if read_size == 0 {
                break;
            }
            inner.offset += read_size;
            total_read_size += read_size;
        }
        total_read_size
    }
    fn write(&self, buf: UserBuffer) -> usize {
        let mut inner = self.inner_exclusive_access();
        let mut total_write_size = 0usize;
        for slice in buf.buffers.iter() {
            let write_size = inner.inode.write().write_at(inner.offset, *slice);
            assert_eq!(write_size, slice.len());
            inner.offset += write_size;
            total_write_size += write_size;
        }
        total_write_size
    }
}

/// 获取根结点
pub fn root_inode() -> Arc<OSInode> {
    Arc::clone(&ROOT_INODE)
}

pub fn list_apps() {
    println!("----------------- ls -----------------");
    let root_osinode = root_inode();
    let root_inode = &root_osinode.inner_exclusive_access().inode;
    for (_, app, _) in root_inode.read().list(0) {
        println!("{}", app);
    }
    println!("--------------------------------------");
}

bitflags! {
    pub struct OpenFlags: u32 {
        const RDONLY = 0;
        const WRONLY = 1 << 0;
        const RDWR = 1 << 1;
        const CREATE = 0x40;
        const O_EXCL = 0x80;
        const O_TRUNC = 0x200;
        const O_APPEND = 0x400;
        const O_NONBLOCK = 0x800;
        const LARGEFILE = 0x8000;
        const O_DIRECTORY = 0x10000;
        const CLOEXEC = 0x80000;
        const O_PATH = 0x200000;
    }
}

impl OpenFlags {
    /// Do not check validity for simplicity
    /// Return (readable, writable)
    pub fn read_write(&self) -> (bool, bool) {
        if self.is_empty() {
            (true, false)
        } else if self.contains(Self::WRONLY) {
            (false, true)
        } else {
            (true, true)
        }
    }
}

/// 打开指定目录下的文件
///
/// 这里有许多加锁动作，且之后会对打开文件表、inode 表进行查找，
/// 查找时会遍历所有的结点，因此也有加锁动作，非常容易出现死锁，需要仔细考虑
pub fn open_file(dir: Arc<OSInode>, name: &str, flags: OpenFlags) -> Option<Arc<OSInode>> {
    // println!("in open_file");
    let dir = &dir.inner_exclusive_access().inode;
    let (readable, writable) = flags.read_write();
    if flags.contains(OpenFlags::CREATE) {
        let read_dir = dir.read();
        if let Some(inode) = read_dir.find_entry(name) {
            drop(read_dir);
            let res = insert_file_table(readable, writable, inode);
            // clear size
            res.clear();
            Some(res)
        } else {
            // create file
            drop(read_dir);
            let mut write_dir = dir.write();
            // create 检查文件不存在会返回 None，而这已经检查过了，可以直接 unwrap
            let res = write_dir.create(name, Attribute::ARCHIVE).unwrap();
            // 注意解锁以便可以遍历文件表
            drop(write_dir);
            Some(insert_file_table(readable, writable, res))
        }
    } else {
        dir.read().find_entry(name).map(|tmp| {
            let inode = insert_file_table(readable, writable, tmp);
            if flags.contains(OpenFlags::O_TRUNC) {
                inode.clear();
            } else if flags.contains(OpenFlags::O_APPEND) {
                inode.seek(inode.size());
            }
            inode
        })
    }
}

/// 根据路径进行搜索，直到达到最顶层目录
///
/// 若为绝对路径，则从根节点开始查找</br>
/// 若为相对路径，且 fd = AT_FDCWD，则从当前工作目录开始查找</br>
/// 若为相对路径，且 fd 为一个合法文件描述符，则从指向的目录开始查找
///
/// 返回顶层目录的 osinode 结点，但该结点为临时结点，并不添加到 file_table 中
fn get_dir(fd: isize, path: &str) -> Option<Arc<OSInode>> {
    // println!("in get_dir");
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let relative_osinode;
    if path.starts_with('/') {
        relative_osinode = root_inode();
    } else {
        if fd == AT_FDCWD {
            relative_osinode = Arc::clone(&task_inner.cwd);
        } else {
            match task_inner.fd_table.get(fd as usize).unwrap() {
                FileClass::OSInode(osinode) => relative_osinode = Arc::clone(&osinode),
                _ => return None,
            }
        }
    }

    let relative_inode = &relative_osinode.inner_exclusive_access().inode;
    let res = relative_inode
        .read()
        .get_dir(path)
        .map(|parent| Arc::new(OSInode::new(true, true, Arc::new(RwLock::new(parent)))));
    // println!("out get_dir");
    res
}

/// 根据路径打开文件
///
/// fd 意义与 get_dir 相同
pub fn open_file_at(fd: isize, path: &str, flags: OpenFlags) -> Option<Arc<OSInode>> {
    // println!("in open_file_at");
    match get_dir(fd, path) {
        Some(parent) => {
            match path
                .trim_end_matches('/')
                .split('/')
                .collect::<Vec<_>>()
                .pop()
            {
                Some(name) => open_file(parent, name, flags),
                None => None,
            }
        }
        None => None,
    }
}

/// 创建目录
///
/// 只创建目录文件，而不打开
pub fn mkdir_at(fd: isize, path: &str) -> isize {
    // println!("in mkdir_at");
    match get_dir(fd, path) {
        Some(parent) => {
            match path
                .trim_end_matches('/')
                .split('/')
                .collect::<Vec<_>>()
                .pop()
            {
                Some(name) => {
                    let parent = &parent.inner_exclusive_access().inode;
                    parent.write().create(name, Attribute::DIRECTORY);
                    // println!("out mkdir_at");
                    0
                }
                None => -1,
            }
        }
        None => -1,
    }
}

/// 移除指定文件的链接(可用于删除文件)
///
/// 目前实现仅能删除文件
pub fn unlinkat(fd: isize, path: &str) -> isize {
    if let Some(inode) = open_file_at(fd, path, OpenFlags::RDONLY) {
        inode.inner_exclusive_access().inode.write().remove();
        drop(inode);
        remove_from_file_table();
        0
    } else {
        -1
    }
}

/// 移动文件
pub fn rename(olddirfd: isize, oldpath: &str, newdirfd: isize, newpath: &str) -> isize {
    let mut buf;
    let is_dir;
    if let Some(osinode) = open_file_at(olddirfd, oldpath, OpenFlags::RDWR) {
        buf = osinode.read_all();
        let inner = osinode.inner_exclusive_access();
        let mut inode = inner.inode.write();
        is_dir = inode.is_dir();
        inode.remove();
        drop(inode);
        drop(inner);
        drop(osinode);
        remove_from_file_table();
    } else {
        return -1;
    }

    if is_dir {
        mkdir_at(newdirfd, newpath);
    }
    if let Some(osinode) = open_file_at(newdirfd, newpath, OpenFlags::CREATE | OpenFlags::WRONLY) {
        osinode.write(UserBuffer::new(vec![unsafe {
            core::slice::from_raw_parts_mut(buf.as_mut_slice().as_mut_ptr(), buf.len())
        }]));
        0
    } else {
        -1
    }
}

pub fn add_initproc_shell() {
    extern "C" {
        fn _num_app();
    }
    let num_app_ptr = _num_app as usize as *mut usize;
    let app_start = unsafe { core::slice::from_raw_parts_mut(num_app_ptr.add(1), 3) };

    if let Some(inode) = open_file(root_inode(), "initproc", OpenFlags::CREATE) {
        println!("Create initproc ");
        let mut data: Vec<&'static mut [u8]> = Vec::new();
        data.push(unsafe {
            core::slice::from_raw_parts_mut(app_start[0] as *mut u8, app_start[1] - app_start[0])
        });
        println!("Start write initproc ");
        inode.write(UserBuffer::new(data));
    } else {
        panic!("initproc create fail!");
    }

    if let Some(inode) = open_file(root_inode(), "user_shell", OpenFlags::CREATE) {
        println!("Create user_shell ");
        let mut data: Vec<&'static mut [u8]> = Vec::new();
        data.push(unsafe {
            core::slice::from_raw_parts_mut(app_start[1] as *mut u8, app_start[2] - app_start[1])
        });
        println!("Start write user_shell ");
        inode.write(UserBuffer::new(data));
    } else {
        panic!("user_shell create fail!");
    }

    // 虽然这两个文件后面要重新打开，但这里强制写回文件防止出错
    remove_from_file_table();
    println!("Write apps(initproc & user_shell) to disk from mem");
}
