#![no_std]
#![no_main]
#![feature(panic_info_message)]
#![feature(alloc_error_handler)]

extern crate alloc;

#[macro_use]
extern crate bitflags;

#[cfg(feature = "board_k210")]
#[path = "boards/k210.rs"]
mod board;
#[cfg(not(any(feature = "board_k210")))]
#[path = "boards/qemu.rs"]
mod board;

#[cfg(feature = "multi-core")]
pub const CORE_NUM: usize = 2;
#[cfg(not(any(feature = "multi-core")))]
pub const CORE_NUM: usize = 1;

#[macro_use]
mod console;
mod config;
mod drivers;
mod fs;
mod lang_items;
mod mm;
mod sbi;
mod signal;
mod syscall;
mod task;
mod timer;
mod trap;
mod types;
mod utsname;

use core::{
    arch::global_asm,
    sync::atomic::{AtomicBool, Ordering},
};

use riscv::register::sstatus::{self, FS};

use crate::task::hartid;

global_asm!(include_str!("entry.asm"));
global_asm!(include_str!("link_app.S"));

fn clear_bss() {
    extern "C" {
        fn sbss();
        fn ebss();
    }
    unsafe {
        core::slice::from_raw_parts_mut(sbss as usize as *mut u8, ebss as usize - sbss as usize)
            .fill(0);
    }
}

static LOCK: AtomicBool = AtomicBool::new(false);

fn set_fs() {
    unsafe {
        sstatus::set_fs(FS::Clean);
    }
}

// 避免编译器对名字进行混淆
#[no_mangle]
pub fn rust_main() -> ! {
    // rustsbi 同时开启了所有核
    // 这里先让一个核启动初始化，其它核自旋
    if hartid() == 0 {
        clear_bss();
        println!("[kernel] init core0");
        println!("[kernel] Hello, world!");
        mm::init();
        mm::activate();
        mm::remap_test();
        set_fs();
        trap::init();
        trap::enable_timer_interrupt();
        timer::set_next_trigger();
        fs::add_initproc_shell();
        fs::list_apps();
        task::add_initproc();
        LOCK.store(true, Ordering::SeqCst);
        task::run_tasks();
    } else {
        while LOCK.load(Ordering::SeqCst) == false {}
        println!("[kernel] init core1");
        mm::activate();
        set_fs();
        trap::init();
        trap::enable_timer_interrupt();
        timer::set_next_trigger();
        task::run_tasks();
    }
    panic!("Unreachable in rust_main!");
}
