use super::{frame_alloc, FrameTracker};
use super::{PTEFlags, PageTable, PageTableEntry};
use super::{PhysAddr, PhysPageNum, VirtAddr, VirtPageNum};
use super::{StepByOne, VPNRange};
use crate::config::*;
use crate::fs::{open_file, root_inode, OpenFlags};
use crate::task::{
    Aux, AT_BASE, AT_CLKTCK, AT_EGID, AT_ENTRY, AT_EUID, AT_FLAGS, AT_GID, AT_HWCAP, AT_NOTELF,
    AT_PAGESZ, AT_PHDR, AT_PHENT, AT_PHNUM, AT_PLATFORM, AT_SECURE, AT_UID,
};
use alloc::collections::BTreeMap;
use alloc::sync::Arc;
use alloc::vec::Vec;
use core::arch::asm;
use lazy_static::*;
use riscv::register::satp;
use spin::{Mutex, MutexGuard};

extern "C" {
    fn stext();
    fn etext();
    fn srodata();
    fn erodata();
    fn sdata();
    fn edata();
    fn sbss_with_stack();
    fn ebss();
    fn ekernel();
    fn strampoline();
}

lazy_static! {
    pub static ref KERNEL_SPACE: Arc<MemorySet> = Arc::new(MemorySet::new_kernel());
}

pub fn kernel_token() -> usize {
    KERNEL_SPACE.token()
}

pub struct MemorySet {
    // 可变的
    inner: Mutex<MemorySetInner>,
}

impl MemorySet {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, MemorySetInner> {
        self.inner.lock()
    }

    pub fn new_bare() -> Self {
        Self {
            inner: Mutex::new(MemorySetInner {
                brk: 0.into(),
                mmap_bottom: 0.into(),
                page_table: PageTable::new(),
                areas: Vec::new(),
            }),
        }
    }

    pub fn token(&self) -> usize {
        self.inner_exclusive_access().page_table.token()
    }
}

pub struct MemorySetInner {
    /// 堆高位地址，堆向高地址方向增长
    pub brk: VirtAddr,
    /// mmap 低位地址，mmap 空间向低地址方向增长
    pub mmap_bottom: VirtAddr,
    pub page_table: PageTable,
    pub areas: Vec<MapArea>,
}

impl MemorySet {
    fn push(&self, mut map_area: MapArea, data: Option<&[u8]>, offset: usize) {
        let mut inner = self.inner_exclusive_access();
        map_area.map(&mut inner.page_table);
        if let Some(data) = data {
            map_area.copy_data(&mut inner.page_table, data, offset);
        }
        inner.areas.push(map_area);
    }

    /// 插入的区域允许包含于某个已分配的空间中，其它情况未实现。
    /// 同时可以用来改变访问权限
    pub fn insert_framed_area(
        &self,
        start_va: VirtAddr,
        end_va: VirtAddr,
        permission: MapPermission,
    ) {
        // println!(
        //     "insert_framed_area: start_va = {:?}, end_va = {:?}",
        //     start_va, end_va
        // );
        let mut inner = self.inner_exclusive_access();
        let start_vpn = start_va.floor();
        let end_vpn = end_va.ceil();
        for (id, area) in inner.areas.iter().enumerate() {
            let l = area.vpn_range.get_start();
            let r = area.vpn_range.get_end();
            // 如果映射的区域之前被分配过，则将该地址空间切割开来，同时更新访问权限
            if l <= start_vpn && r >= end_vpn {
                if l == start_vpn && r == end_vpn && permission == area.map_perm {
                    return;
                }
                let area = inner.areas.remove(id);
                let split_areas =
                    split(&mut inner.page_table, area, start_vpn, end_vpn, permission);
                for split_area in split_areas {
                    inner.areas.push(split_area);
                }
                return;
            }
        }
        drop(inner);
        self.push(
            MapArea::new(start_va, end_va, MapType::Framed, permission),
            None,
            0,
        );
    }

    /// 寻找可以插入一个逻辑段的虚拟地址，返回插入的起始地址
    ///
    /// `hint` 意义同 `insert_framed_area_randomly`
    fn find_insert_addr(&self, size: usize, hint: Option<VirtAddr>) -> VirtAddr {
        let inner = self.inner_exclusive_access();
        let num = VirtAddr::from(size).ceil();
        let end_vpn = if hint.is_some() {
            hint.unwrap().ceil()
        } else {
            // 额外减掉一个空白保护页
            VirtPageNum::from(inner.mmap_bottom) - 1
        };
        let start_vpn = end_vpn - num;
        for area in inner.areas.iter() {
            let l = area.vpn_range.get_start();
            let r = area.vpn_range.get_end();
            // 逻辑段冲突，注意逻辑段是左闭右开
            // 重新寻找插入地址时，减去一个保护页的空间（主要针对多线程用户栈）
            if end_vpn > l && start_vpn < r {
                drop(inner);
                return self.find_insert_addr(size, Some((start_vpn - 1).into()));
            }
        }
        start_vpn.into()
    }

    /// 向地址空间随机插入一个逻辑段，返回插入的地址
    ///
    /// 若 `hint` 存在，将其作为一个参考的结束地址，否则，将 `mmap_bottom` 作为一个参考的结束地址，
    /// 在其附近查找可以插入的地址。如果出现冲突，向低地址方向继续查找可插入的位置。
    pub fn insert_framed_area_randomly(
        &self,
        size: usize,
        hint: Option<VirtAddr>,
        permission: MapPermission,
    ) -> (VirtAddr, VirtAddr) {
        let start_va = self.find_insert_addr(size, hint);
        let end_va = (start_va + size).ceil().into();
        self.insert_framed_area(start_va, end_va, permission);
        (start_va, end_va)
    }

    /// 复制逻辑段内容
    pub fn clone_area(&self, start_vpn: VirtPageNum, another: &MemorySet) {
        let another_inner = another.inner_exclusive_access();
        if let Some(area) = another_inner
            .areas
            .iter()
            .find(|area| area.vpn_range.get_start() == start_vpn)
        {
            for vpn in area.vpn_range {
                let src_ppn = another_inner.page_table.translate(vpn).unwrap().ppn();
                let dst_ppn = self.translate(vpn).unwrap().ppn();
                dst_ppn
                    .get_bytes_array()
                    .copy_from_slice(src_ppn.get_bytes_array());
            }
        }
    }

    pub fn remove_area_with_start_vpn(&self, start_vpn: VirtPageNum) {
        let mut inner = self.inner_exclusive_access();
        if let Some((idx, _)) = inner
            .areas
            .iter_mut()
            .enumerate()
            .find(|(_, area)| area.vpn_range.get_start() == start_vpn)
        {
            let mut area = inner.areas.remove(idx);
            area.unmap(&mut inner.page_table);
        }
    }

    pub fn activate(&self) {
        let inner = self.inner_exclusive_access();
        let satp = inner.page_table.token();
        unsafe {
            satp::write(satp);
            asm!("sfence.vma");
        }
    }

    pub fn translate(&self, vpn: VirtPageNum) -> Option<PageTableEntry> {
        let inner = self.inner_exclusive_access();
        inner.page_table.translate(vpn)
    }

    pub fn recycle_data_pages(&self) {
        let mut inner = self.inner_exclusive_access();
        inner.areas.clear();
    }
}

impl MemorySet {
    /// Mention that trampoline is not collected by areas.
    fn map_trampoline(&self) {
        let mut inner = self.inner_exclusive_access();
        inner.page_table.map(
            VirtAddr::from(TRAMPOLINE).into(),
            PhysAddr::from(strampoline as usize).into(),
            PTEFlags::R | PTEFlags::X,
        );
    }

    /// Without kernel stacks.
    pub fn new_kernel() -> Self {
        let memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        // map kernel sections
        println!(".text [{:#x}, {:#x})", stext as usize, etext as usize);
        println!(".rodata [{:#x}, {:#x})", srodata as usize, erodata as usize);
        println!(".data [{:#x}, {:#x})", sdata as usize, edata as usize);
        println!(
            ".bss [{:#x}, {:#x})",
            sbss_with_stack as usize, ebss as usize
        );
        println!("mapping .text section");
        memory_set.push(
            MapArea::new(
                (stext as usize).into(),
                (etext as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::X,
            ),
            None,
            0,
        );
        println!("mapping .rodata section");
        memory_set.push(
            MapArea::new(
                (srodata as usize).into(),
                (erodata as usize).into(),
                MapType::Identical,
                MapPermission::R,
            ),
            None,
            0,
        );
        println!("mapping .data section");
        memory_set.push(
            MapArea::new(
                (sdata as usize).into(),
                (edata as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
            0,
        );
        println!("mapping .bss section");
        memory_set.push(
            MapArea::new(
                (sbss_with_stack as usize).into(),
                (ebss as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
            0,
        );
        println!("mapping physical memory");
        memory_set.push(
            MapArea::new(
                (ekernel as usize).into(),
                MEMORY_END.into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
            0,
        );
        println!("mapping memory-mapped registers");
        for pair in MMIO {
            memory_set.push(
                MapArea::new(
                    (*pair).0.into(),
                    ((*pair).0 + (*pair).1).into(),
                    MapType::Identical,
                    MapPermission::R | MapPermission::W,
                ),
                None,
                0,
            );
        }
        memory_set
    }

    /// 加载动态链接器
    fn load_dynamic_linker(&self) -> usize {
        // println!("in load_dynamic_linker");
        if let Some(app_inode) = open_file(root_inode(), "libc.so", OpenFlags::RDONLY) {
            let app_data = app_inode.read_all();
            let elf_data = app_data.as_slice();
            let elf = xmas_elf::ElfFile::new(elf_data).unwrap();
            let elf_header = elf.header;
            let magic = elf_header.pt1.magic;
            assert_eq!(magic, [0x7f, 0x45, 0x4c, 0x46], "invalid elf!");
            let ph_count = elf_header.pt2.ph_count();
            for i in 0..ph_count {
                let ph = elf.program_header(i).unwrap();
                if ph.get_type().unwrap() == xmas_elf::program::Type::Load {
                    let start_va: VirtAddr =
                        (DYNAMIC_LINKER_BASE + ph.virtual_addr() as usize).into();
                    let end_va: VirtAddr =
                        (DYNAMIC_LINKER_BASE + (ph.virtual_addr() + ph.mem_size()) as usize).into();
                    let mut map_perm = MapPermission::U;
                    let ph_flags = ph.flags();
                    if ph_flags.is_read() {
                        map_perm |= MapPermission::R;
                    }
                    if ph_flags.is_write() {
                        map_perm |= MapPermission::W;
                    }
                    if ph_flags.is_execute() {
                        map_perm |= MapPermission::X;
                    }
                    let map_area = MapArea::new(start_va, end_va, MapType::Framed, map_perm);
                    // println!(
                    //     "add map area: start = {:#x}, end = {:#x}, perm = {:?}",
                    //     start_va.0, end_va.0, map_perm
                    // );
                    self.push(
                        map_area,
                        Some(
                            &elf.input
                                [ph.offset() as usize..(ph.offset() + ph.file_size()) as usize],
                        ),
                        start_va.page_offset(),
                    );
                }
            }
            // println!("success load dynamic linker");
            elf_header.pt2.entry_point() as usize
        } else {
            panic!("fail to load dynamic linker");
        }
    }

    /// 解析 `ELF` 文件并生成用户对应的地址空间，包括跳板页、动态链接（未完成），
    /// 不包括 `trap` 上下文以及用户栈
    ///
    /// 返回值: (memory_set, entry_point, Vec\<Aux\>)
    pub fn from_elf(elf_data: &[u8]) -> (Self, usize, Vec<Aux>) {
        let mut auxv = Vec::new();
        let memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        // map program headers of elf, with U flag
        let elf = xmas_elf::ElfFile::new(elf_data).unwrap();
        let elf_header = elf.header;
        let magic = elf_header.pt1.magic;
        assert_eq!(magic, [0x7f, 0x45, 0x4c, 0x46], "invalid elf!");
        let ph_count = elf_header.pt2.ph_count();

        // 检查是否是动态链接并加载
        let mut interp_base = 0;
        let mut entry = elf_header.pt2.entry_point() as usize;
        for i in 0..ph_count {
            let ph = elf.program_header(i).unwrap();
            if ph.get_type().unwrap() == xmas_elf::program::Type::Interp {
                interp_base = DYNAMIC_LINKER_BASE;
                entry = memory_set.load_dynamic_linker() + DYNAMIC_LINKER_BASE;
                break;
            }
        }

        let mut max_end_vpn = VirtPageNum(0);
        let mut head_va = 0;
        for i in 0..ph_count {
            let ph = elf.program_header(i).unwrap();
            if ph.get_type().unwrap() == xmas_elf::program::Type::Load {
                let start_va: VirtAddr = (ph.virtual_addr() as usize).into();
                let end_va: VirtAddr = ((ph.virtual_addr() + ph.mem_size()) as usize).into();
                if head_va == 0 {
                    head_va = start_va.0;
                }
                let mut map_perm = MapPermission::U;
                let ph_flags = ph.flags();
                if ph_flags.is_read() {
                    map_perm |= MapPermission::R;
                }
                if ph_flags.is_write() {
                    map_perm |= MapPermission::W;
                }
                if ph_flags.is_execute() {
                    map_perm |= MapPermission::X;
                }
                let map_area = MapArea::new(start_va, end_va, MapType::Framed, map_perm);
                // println!(
                //     "add map area: start = {:#x}, end = {:#x}, perm = {:?}",
                //     start_va.0, end_va.0, map_perm
                // );
                max_end_vpn = map_area.vpn_range.get_end();
                memory_set.push(
                    map_area,
                    Some(&elf.input[ph.offset() as usize..(ph.offset() + ph.file_size()) as usize]),
                    start_va.page_offset(),
                );
            }
        }
        // println!("load section");

        auxv.push(Aux::new(
            AT_PHDR,
            head_va + elf_header.pt2.ph_offset() as usize,
        ));
        auxv.push(Aux::new(AT_PHENT, elf_header.pt2.ph_entry_size() as usize));
        auxv.push(Aux::new(AT_PHNUM, ph_count as usize));
        auxv.push(Aux::new(AT_PAGESZ, PAGE_SIZE));
        auxv.push(Aux::new(AT_BASE, interp_base));
        auxv.push(Aux::new(AT_FLAGS, 0));
        auxv.push(Aux::new(AT_ENTRY, entry as usize));
        auxv.push(Aux::new(AT_NOTELF, 0x112d));
        auxv.push(Aux::new(AT_UID, 0));
        auxv.push(Aux::new(AT_EUID, 0));
        auxv.push(Aux::new(AT_GID, 0));
        auxv.push(Aux::new(AT_EGID, 0));
        auxv.push(Aux::new(AT_PLATFORM, 0));
        auxv.push(Aux::new(AT_HWCAP, 0));
        auxv.push(Aux::new(AT_CLKTCK, 100));
        auxv.push(Aux::new(AT_SECURE, 0));

        let mut inner = memory_set.inner_exclusive_access();
        inner.brk = max_end_vpn.into();
        inner.mmap_bottom = VirtAddr::from(MMAP_TOP);
        drop(inner);

        (memory_set, entry, auxv)
    }

    /// 复制用户地址空间，不包括用户栈以及 `trap` 上下文
    pub fn from_existed_user(user_space: &MemorySet) -> MemorySet {
        let user_space_inner = user_space.inner_exclusive_access();
        let memory_set = Self::new_bare();
        // 映射 trampoline
        memory_set.map_trampoline();
        // 复制段
        for area in user_space_inner.areas.iter() {
            // 检测是否是用户栈或者 trap 上下文空间，处于这些位置的逻辑段不会被复制
            // 需要进程重新分配
            let start_va = VirtAddr::from(area.vpn_range.get_start());
            let end_va = VirtAddr::from(area.vpn_range.get_end());
            if (start_va >= MMAP_TOP.into() && end_va <= USTACK_TOP.into())
                || (start_va >= HIGH_ADDR_SPACE_BOTTOM.into() && end_va <= TRAP_CONTEXT_TOP.into())
            {
                continue;
            }

            let new_area = MapArea::from_another(area);
            memory_set.push(new_area, None, 0);
            for vpn in area.vpn_range {
                // 不能直接调用 self.translate()，会造成死锁
                let src_ppn = user_space_inner.page_table.translate(vpn).unwrap().ppn();
                let dst_ppn = memory_set.translate(vpn).unwrap().ppn();
                dst_ppn
                    .get_bytes_array()
                    .copy_from_slice(src_ppn.get_bytes_array());
            }
        }
        let mut inner = memory_set.inner_exclusive_access();
        inner.brk = user_space_inner.brk;
        inner.mmap_bottom = user_space_inner.mmap_bottom;
        drop(inner);
        memory_set
    }
}

pub struct MapArea {
    // 左闭右开
    vpn_range: VPNRange,
    data_frames: BTreeMap<VirtPageNum, FrameTracker>,
    map_type: MapType,
    map_perm: MapPermission,
}

impl MapArea {
    pub fn new(
        start_va: VirtAddr,
        end_va: VirtAddr,
        map_type: MapType,
        map_perm: MapPermission,
    ) -> Self {
        let start_vpn: VirtPageNum = start_va.floor();
        let end_vpn: VirtPageNum = end_va.ceil();
        Self {
            vpn_range: VPNRange::new(start_vpn, end_vpn),
            data_frames: BTreeMap::new(),
            map_type,
            map_perm,
        }
    }
    pub fn from_another(another: &MapArea) -> Self {
        Self {
            vpn_range: VPNRange::new(another.vpn_range.get_start(), another.vpn_range.get_end()),
            data_frames: BTreeMap::new(),
            map_type: another.map_type,
            map_perm: another.map_perm,
        }
    }
    pub fn map_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        let ppn: PhysPageNum;
        match self.map_type {
            MapType::Identical => {
                ppn = PhysPageNum(vpn.0);
            }
            MapType::Framed => {
                let frame = frame_alloc().unwrap();
                ppn = frame.ppn;
                self.data_frames.insert(vpn, frame);
            }
        }
        let pte_flags = PTEFlags::from_bits(self.map_perm.bits).unwrap();
        page_table.map(vpn, ppn, pte_flags);
    }
    pub fn unmap_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        if self.map_type == MapType::Framed {
            self.data_frames.remove(&vpn);
        }
        page_table.unmap(vpn);
    }
    pub fn map(&mut self, page_table: &mut PageTable) {
        for vpn in self.vpn_range {
            // println!("map vpn = {:?}", vpn);
            self.map_one(page_table, vpn);
        }
    }
    pub fn unmap(&mut self, page_table: &mut PageTable) {
        for vpn in self.vpn_range {
            self.unmap_one(page_table, vpn);
        }
    }

    /// 将 `data` 复制到该逻辑段偏移 `offset` 的地方
    ///
    /// assume that all frames were cleared before
    pub fn copy_data(&mut self, page_table: &mut PageTable, data: &[u8], offset: usize) {
        assert_eq!(self.map_type, MapType::Framed);
        let mut start: usize = 0;
        let mut current_vpn = self.vpn_range.get_start();
        let len = data.len();

        // 页表数据不对齐，先处理不对齐的部分
        let src = &data[..len.min(PAGE_SIZE - offset)];
        let dst = &mut page_table
            .translate(current_vpn)
            .unwrap()
            .ppn()
            .get_bytes_array()[offset..offset + src.len()];
        dst.copy_from_slice(src);
        start += PAGE_SIZE - offset;
        current_vpn.step();

        loop {
            if start >= len {
                break;
            }
            let src = &data[start..len.min(start + PAGE_SIZE)];
            let dst = &mut page_table
                .translate(current_vpn)
                .unwrap()
                .ppn()
                .get_bytes_array()[..src.len()];
            dst.copy_from_slice(src);
            start += PAGE_SIZE;
            current_vpn.step();
        }
    }
}

/// 拆分逻辑段
///
/// `[start, end) -> [start, left), [left, right), [right, end)`
pub fn split(
    page_table: &mut PageTable,
    mut area: MapArea,
    left: VirtPageNum,
    right: VirtPageNum,
    map_perm: MapPermission,
) -> Vec<MapArea> {
    let mut res = Vec::new();
    let start = area.vpn_range.get_start();
    let end = area.vpn_range.get_end();

    if start < left {
        let mut left_area = MapArea::new(start.into(), left.into(), area.map_type, area.map_perm);
        let mut cur = start;
        while cur < left {
            left_area
                .data_frames
                .insert(cur, area.data_frames.remove(&cur).unwrap());
            cur.step();
        }
        res.push(left_area);
        // println!(
        //     "split MapArea: start_va = {:?}, end_va = {:?}, permission = {:?}",
        //     VirtAddr::from(start),
        //     VirtAddr::from(left),
        //     area.map_perm,
        // );
    }

    let mut mid_area = MapArea::new(left.into(), right.into(), area.map_type, map_perm);
    let mut cur = left;
    while cur < right {
        mid_area
            .data_frames
            .insert(cur, area.data_frames.remove(&cur).unwrap());
        page_table.set_flags(cur, PTEFlags::from_bits(map_perm.bits).unwrap());
        cur.step();
    }
    res.push(mid_area);
    // println!(
    //     "split MapArea: start_va = {:?}, end_va = {:?}, permission = {:?}",
    //     VirtAddr::from(left),
    //     VirtAddr::from(right),
    //     map_perm,
    // );

    if end > right {
        let mut right_area = MapArea::new(right.into(), end.into(), area.map_type, area.map_perm);
        let mut cur = right;
        while cur < end {
            right_area
                .data_frames
                .insert(cur, area.data_frames.remove(&cur).unwrap());
            cur.step();
        }
        res.push(right_area);
        // println!(
        //     "split MapArea: start_va = {:?}, end_va = {:?}",
        //     VirtAddr::from(right),
        //     VirtAddr::from(end)
        // );
    }
    res
}

#[derive(Copy, Clone, PartialEq, Debug)]
pub enum MapType {
    Identical,
    Framed,
}

bitflags! {
    pub struct MapPermission: u8 {
        const R = 1 << 1;
        const W = 1 << 2;
        const X = 1 << 3;
        const U = 1 << 4;
    }
}

#[allow(unused)]
pub fn remap_test() {
    let mut kernel_space = KERNEL_SPACE.inner_exclusive_access();
    let mid_text: VirtAddr = ((stext as usize + etext as usize) / 2).into();
    let mid_rodata: VirtAddr = ((srodata as usize + erodata as usize) / 2).into();
    let mid_data: VirtAddr = ((sdata as usize + edata as usize) / 2).into();
    assert!(!kernel_space
        .page_table
        .translate(mid_text.floor())
        .unwrap()
        .writable(),);
    assert!(!kernel_space
        .page_table
        .translate(mid_rodata.floor())
        .unwrap()
        .writable(),);
    assert!(!kernel_space
        .page_table
        .translate(mid_data.floor())
        .unwrap()
        .executable(),);
    println!("remap_test passed!");
}
