use alloc::sync::Arc;
use alloc::vec;

use crate::fs::{
    make_pipe, mkdir_at, open_file_at, remove_from_file_table, rename, unlinkat, File, FileClass,
    IOVec, Kstat, OpenFlags, Winsize, AT_FDCWD,
};
use crate::mm::{
    translated_byte_buffer, translated_ref, translated_refmut, translated_str, UserBuffer,
};
use crate::syscall::errno::ENOENT;
use crate::task::{current_task, current_user_token, suspend_current_and_run_next};
use crate::timer::{current_timespec, Timespec};
use crate::types::FdSet;

pub fn sys_getcwd(buf: *mut u8, size: usize) -> isize {
    let token = current_user_token();
    let mut buffer = UserBuffer::new(translated_byte_buffer(token, buf, size));
    buffer.read(
        current_task()
            .unwrap()
            .inner_exclusive_access()
            .cwd
            .inner_exclusive_access()
            .inode
            .read()
            .path()
            .as_bytes(),
    );
    buf as isize
}

pub fn sys_write(fd: usize, buf: *const u8, len: usize) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let f: Arc<dyn File + Send + Sync> = match file {
            FileClass::OSInode(f) => f,
            FileClass::PIPE(f) => f,
            FileClass::Other(f) => f,
        };
        // 文件读写会挂起当前线程，需要释放锁
        drop(task_inner);
        drop(task);
        if !f.writable() {
            return -1;
        }
        f.write(UserBuffer::new(translated_byte_buffer(token, buf, len))) as isize
    } else {
        -1
    }
}

pub fn sys_read(fd: usize, buf: *const u8, len: usize) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let f: Arc<dyn File + Send + Sync> = match file {
            FileClass::OSInode(f) => f,
            FileClass::PIPE(f) => f,
            FileClass::Other(f) => f,
        };
        // 文件读写会挂起当前线程，需要释放锁
        drop(task_inner);
        drop(task);
        if !f.readable() {
            return -1;
        }
        f.read(UserBuffer::new(translated_byte_buffer(token, buf, len))) as isize
    } else {
        -1
    }
}

const SEEK_SET: usize = 0;
const SEEK_CUR: usize = 1;
const SEEK_END: usize = 2;

pub fn sys_lseek(fd: usize, offset: usize, whence: usize) -> isize {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        match file {
            FileClass::OSInode(inode) => {
                let new_offset = match whence {
                    SEEK_SET => offset,
                    SEEK_CUR => inode.offset() + offset,
                    SEEK_END => inode.size() + offset,
                    _ => usize::MAX as usize,
                };
                inode.seek(new_offset);
                new_offset as isize
            }
            _ => -1,
        }
    } else {
        -1
    }
}

pub fn sys_openat(dirfd: isize, path: *const u8, flags: u32, _mode: u32) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    // println!("in sys_openat: dirfd = {}, path = {}, flags = {:#x}", dirfd, path, flags);
    let flags = OpenFlags::from_bits(flags).unwrap();
    // println!(
    //     "in sys_openat: dirfd = {}, path = {:?}, flags = {:?}",
    //     dirfd, path, flags
    // );
    if let Some(inode) = open_file_at(dirfd, path.as_str(), flags) {
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();
        let fd = task_inner.alloc_fd();
        task_inner.fd_table.set_fd(fd, FileClass::OSInode(inode));
        fd as isize
    } else {
        -1
    }
}

pub fn sys_mkdirat(dirfd: isize, path: *const u8, _mode: u32) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    mkdir_at(dirfd, path.as_str())
}

pub fn sys_chdir(path: *const u8) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);

    // 在 open_file_at 中会用到当前进程变量，因此会获得锁，
    // 如果把这段放后面会死锁
    let new_cwd = open_file_at(AT_FDCWD, &path, OpenFlags::RDWR).unwrap();

    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    task_inner.cwd = new_cwd;
    remove_from_file_table();
    0
}

pub fn sys_close(fd: usize) -> isize {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if task_inner.fd_table.get(fd).is_some() {
        task_inner.fd_table.remove(fd);
        remove_from_file_table();
        0
    } else {
        -1
    }
}

pub fn sys_getdents64(fd: usize, buf: *const u8, len: usize) -> isize {
    let token = current_user_token();
    let mut buf = UserBuffer::new(translated_byte_buffer(token, buf, len));
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        match file {
            FileClass::OSInode(inode) => {
                return buf.read(inode.getdents64().as_slice()) as isize;
            }
            _ => return -1,
        }
    } else {
        return -1;
    }
}

pub fn sys_readv(fd: usize, iov: *const IOVec, iovcnt: usize) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let f: Arc<dyn File + Send + Sync> = match file {
            FileClass::OSInode(f) => f.clone(),
            FileClass::PIPE(f) => f.clone(),
            FileClass::Other(f) => f.clone(),
        };
        let mut size = 0;
        for i in 0..iovcnt {
            let iovec = translated_ref(token, unsafe { iov.add(i) });
            let buf = translated_byte_buffer(token, iovec.iov_base as *const u8, iovec.iov_len);
            size += f.read(UserBuffer::new(buf));
        }
        return size as isize;
    } else {
        -1
    }
}

pub fn sys_writev(fd: usize, iov: *const IOVec, iovcnt: usize) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let f: Arc<dyn File + Send + Sync> = match file {
            FileClass::OSInode(f) => f.clone(),
            FileClass::PIPE(f) => f.clone(),
            FileClass::Other(f) => f.clone(),
        };
        let mut size = 0;
        for i in 0..iovcnt {
            let iovec = translated_ref(token, unsafe { iov.add(i) });
            let buf = translated_byte_buffer(token, iovec.iov_base as *const u8, iovec.iov_len);
            size += f.write(UserBuffer::new(buf));
        }
        return size as isize;
    } else {
        -1
    }
}

pub fn sys_sendfile(out_fd: usize, in_fd: usize, ptr: *mut usize, count: usize) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    // 内存不足，所以申请的缓冲大小只有 512B，循环读取
    let mut buf = vec![0u8; 512];
    let mut len = 512;
    let mut size = 0;

    let fout: Arc<dyn File + Send + Sync>;
    if let Some(file) = task_inner.fd_table.get(out_fd) {
        fout = match file {
            FileClass::OSInode(f) => f.clone(),
            FileClass::PIPE(f) => f.clone(),
            FileClass::Other(f) => f.clone(),
        };
    } else {
        return -1;
    };

    if let Some(file) = task_inner.fd_table.get(in_fd) {
        match file {
            FileClass::OSInode(fin) => {
                let offset = fin.offset();
                if ptr as usize != 0 {
                    fin.seek(*translated_ref(token, ptr));
                }

                while size < count && len == 512 {
                    let buf_read = UserBuffer::new(vec![unsafe {
                        core::slice::from_raw_parts_mut(buf.as_mut_slice().as_mut_ptr(), 512)
                    }]);
                    len = fin.read(buf_read);

                    let buf_write = UserBuffer::new(vec![unsafe {
                        core::slice::from_raw_parts_mut(buf.as_mut_slice().as_mut_ptr(), len)
                    }]);
                    fout.write(buf_write) as isize;

                    size += len;
                }

                if ptr as usize != 0 {
                    *translated_refmut(token, ptr) = fin.offset();
                    fin.seek(offset);
                }
            }
            _ => return -1,
        };
    } else {
        return -1;
    }

    size as isize
}

pub fn sys_pselect6(
    nfds: usize,
    readfds: *mut FdSet,
    writefds: *mut FdSet,
    exceptfds: *mut FdSet,
    timeout: *const Timespec,
    _sigmash: usize,
) -> isize {
    let token = current_user_token();
    let check_readfds = if readfds as usize == 0 {
        FdSet(0)
    } else {
        *translated_ref(token, readfds)
    };
    let check_writefds = if writefds as usize == 0 {
        FdSet(0)
    } else {
        *translated_ref(token, writefds)
    };
    let block = timeout as usize == 0;
    let timeout = if timeout as usize == 0 {
        Timespec {
            tv_sec: 0,
            tv_nsec: 0,
        }
    } else {
        current_timespec() + *(translated_ref(token, timeout))
    };
    println!("in sys_pselect6: timeout = {:#x?}", timeout);

    let mut res = 0;
    let mut kernel_readfds = FdSet(0);
    let mut kernel_writefds = FdSet(0);
    loop {
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();
        for fd in 0..nfds {
            if check_readfds.is_set(fd) && !kernel_readfds.is_set(fd) {
                if let Some(file) = task_inner.fd_table.get(fd) {
                    match file {
                        FileClass::PIPE(f) => {
                            println!("check pipe: fd = {}", fd);
                            if f.available_read() {
                                kernel_readfds.set_fd(fd);
                                res += 1;
                            }
                        }
                        _ => continue,
                    }
                }
            }

            if check_writefds.is_set(fd) && !kernel_writefds.is_set(fd) {
                if let Some(file) = task_inner.fd_table.get(fd) {
                    match file {
                        FileClass::PIPE(f) => {
                            if f.available_write() {
                                kernel_writefds.set_fd(fd);
                                res += 1;
                            }
                        }
                        _ => continue,
                    }
                }
            }
        }

        let cur_timespec = current_timespec();
        if (block && res > 0)
            || (!block
                && cur_timespec.tv_sec >= timeout.tv_sec
                && cur_timespec.tv_nsec >= timeout.tv_nsec)
        {
            break;
        } else {
            drop(task_inner);
            drop(task);
            suspend_current_and_run_next();
        }
    }

    if readfds as usize != 0 {
        *translated_refmut(token, readfds) = kernel_readfds;
    }
    if writefds as usize != 0 {
        *translated_refmut(token, writefds) = kernel_writefds;
    }
    if exceptfds as usize != 0 {
        *translated_refmut(token, exceptfds) = FdSet(0);
    }
    res
}

pub fn sys_readlinkat(dirfd: isize, path: *const u8, buf: *const u8, size: u32) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    let mut buf = UserBuffer::new(translated_byte_buffer(token, buf, size as usize));
    println!("in sys_readlinkat: dirfd = {}, path = {:?}", dirfd, path);
    if path != "/proc/self/exe" {
        unimplemented!();
    }
    buf.read("/".as_bytes()) as isize
}

pub fn sys_fstat(fd: usize, buf: *const u8) -> isize {
    // println!("in sys_fstat: fd = {}", fd as isize);
    let token = current_user_token();
    let mut buf = UserBuffer::new(translated_byte_buffer(
        token,
        buf,
        core::mem::size_of::<Kstat>(),
    ));
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        match file {
            FileClass::OSInode(inode) => {
                buf.read(inode.fstat().as_bytes());
                return 0;
            }
            _ => return -1,
        }
    } else {
        return -1;
    }
}

pub fn sys_fstatat(dirfd: isize, path: *const u8, buf: *const u8, _flags: usize) -> isize {
    let token = current_user_token();
    let mut buf = UserBuffer::new(translated_byte_buffer(
        token,
        buf,
        core::mem::size_of::<Kstat>(),
    ));
    let path = translated_str(token, path);
    // println!("in sys_fstatat: dirfd = {}, path = {:?}", dirfd, path);
    if let Some(inode) = open_file_at(dirfd, path.as_str(), OpenFlags::RDONLY) {
        buf.read(inode.fstat().as_bytes());
        0
    } else {
        -ENOENT
    }
}

pub fn sys_utimensat(
    dirfd: isize,
    path: *const u8,
    _times: *const Timespec,
    _flags: usize,
) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    let flags = OpenFlags::CREATE | OpenFlags::RDWR;
    // println!("in sys_utimensat: dirfd = {}, path = {:?}", dirfd, path);
    if open_file_at(dirfd, path.as_str(), flags).is_some() {
        0
    } else {
        -1
    }
}

pub fn sys_pipe2(pipe: *mut u32) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let (pipe_read, pipe_write) = make_pipe();
    let read_fd = task_inner.alloc_fd();
    task_inner
        .fd_table
        .set_fd(read_fd, FileClass::PIPE(pipe_read));
    let write_fd = task_inner.alloc_fd();
    task_inner
        .fd_table
        .set_fd(write_fd, FileClass::PIPE(pipe_write));
    // println!("in sys_pipe2: read_fd = {}, write_fd = {}", read_fd, write_fd);
    *translated_refmut(token, pipe) = read_fd as u32;
    *translated_refmut(token, unsafe { pipe.add(1) }) = write_fd as u32;
    0
}

pub fn sys_dup(fd: usize) -> isize {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let new_fd = task_inner.alloc_fd();
        task_inner.fd_table.set_fd(new_fd, file);
        new_fd as isize
    } else {
        -1
    }
}

pub fn sys_dup3(oldfd: usize, newfd: usize) -> isize {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if oldfd == newfd {
        return oldfd as isize;
    }
    if let Some(file) = task_inner.fd_table.get(oldfd) {
        task_inner.fd_table.set_fd(newfd, file);
        newfd as isize
    } else {
        -1
    }
}

const F_DUPFD: usize = 0;
const F_GETFD: usize = 1;
const F_SETFD: usize = 2;

pub fn sys_fcntl(fd: usize, cmd: usize, _arg: usize) -> isize {
    // println!("in sys_fcntl: cmd = {}", cmd);
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        match cmd {
            F_DUPFD => {
                let newfd = task_inner.alloc_fd();
                task_inner.fd_table.set_fd(newfd, file);
                newfd as isize
            }
            F_GETFD | F_SETFD => 0,
            _ => 0,
        }
    } else {
        -1
    }
}

pub fn sys_ioctl(_fd: usize, _request: usize, argp: usize) -> isize {
    // println!("in sys_ioctl: fd = {}, request = {:#x}", fd, _request);
    let token = current_user_token();
    *translated_refmut(token, argp as *mut Winsize) = Winsize::new();
    0
}

pub fn sys_unlinkat(dirfd: isize, path: *const u8, _flags: u32) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    // println!("in sys_unlinkat: dirfd = {}, path = {:?}", dirfd, path);
    unlinkat(dirfd, path.as_str())
}

pub fn sys_faccessat(dirfd: isize, path: *const u8, _mode: u32, _flags: usize) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    // println!("in sys_faccessat: dirfd = {}, path = {:?}", dirfd, path);
    if open_file_at(dirfd, path.as_str(), OpenFlags::RDWR).is_some() {
        0
    } else {
        -1
    }
}

pub fn sys_renameat2(
    olddirfd: isize,
    oldpath: *const u8,
    newdirfd: isize,
    newpath: *const u8,
    _flags: u32,
) -> isize {
    let token = current_user_token();
    let oldpath = translated_str(token, oldpath);
    let newpath = translated_str(token, newpath);
    rename(olddirfd, &oldpath, newdirfd, &newpath)
}

pub fn sys_mount(
    _special: *const u8,
    _dir: *const u8,
    _fstype: *const u8,
    _flags: usize,
    _data: *const u8,
) -> isize {
    0
}

pub fn sys_umount(_special: *const u8, _flags: usize) -> isize {
    0
}
