const SYSCALL_GETCWD: usize = 17;
const SYSCALL_DUP: usize = 23;
const SYSCALL_DUP3: usize = 24;
const SYSCALL_FCNTL: usize = 25;
const SYSCALL_IOCTL: usize = 29;
const SYSCALL_MKDIRAT: usize = 34;
const SYSCALL_UNLINKAT: usize = 35;
const SYSCALL_UMOUNT2: usize = 39;
const SYSCALL_MOUNT: usize = 40;
const SYSCALL_FACCESSAT: usize = 48;
const SYSCALL_CHDIR: usize = 49;
const SYSCALL_OPENAT: usize = 56;
const SYSCALL_CLOSE: usize = 57;
const SYSCALL_PIPE: usize = 59;
const SYSCALL_GETDENTS64: usize = 61;
const SYSCALL_LSEEK: usize = 62;
const SYSCALL_READ: usize = 63;
const SYSCALL_WRITE: usize = 64;
const SYSCALL_READV: usize = 65;
const SYSCALL_WRITEV: usize = 66;
const SYSCALL_SENDFILE: usize = 71;
const SYSCALL_PSELECT6: usize = 72;
const SYSCALL_READLINKAT: usize = 78;
const SYSCALL_FSTATAT: usize = 79;
const SYSCALL_FSTAT: usize = 80;
const SYSCALL_UTIMENSAT: usize = 88;
const SYSCALL_EXIT: usize = 93;
const SYSCALL_EXIT_GROUP: usize = 94;
const SYSCALL_SET_TID_ADDRESS: usize = 96;
const SYSCALL_NANOSLEEP: usize = 101;
const SYSCALL_CLOCK_GETTIME: usize = 113;
const SYSCALL_SYSLOG: usize = 116;
const SYSCALL_SCHED_YIELD: usize = 124;
const SYSCALL_KILL: usize = 129;
const SYSCALL_RT_SIGACTION: usize = 134;
const SYSCALL_RT_SIGPROCMASK: usize = 135;
const SYSCALL_RT_SIGTIMEDWAIT: usize = 137;
pub const SYSCALL_RT_SIGRETURN: usize = 139;
const SYSCALL_TIMES: usize = 153;
const SYSCALL_UNAME: usize = 160;
const SYSCALL_GETRUSAGE: usize = 165;
const SYSCALL_UMASK: usize = 166;
const SYSCALL_GET_TIME_OF_DAY: usize = 169;
const SYSCALL_GETPID: usize = 172;
const SYSCALL_GETPPID: usize = 173;
const SYSCALL_GETUID: usize = 174;
const SYSCALL_GETTID: usize = 178;
const SYSCALL_SYSINFO: usize = 179;
const SYSCALL_BRK: usize = 214;
const SYSCALL_MUNMAP: usize = 215;
const SYSCALL_CLONE: usize = 220;
const SYSCALL_EXECVE: usize = 221;
const SYSCALL_MMAP: usize = 222;
const SYSCALL_MPROTECT: usize = 226;
const SYSCALL_WAIT4: usize = 260;
const SYSCALL_PRLIMIT: usize = 261;
const SYSCALL_RENAMEAT2: usize = 276;
const SYSCALL_SHUTDOWN: usize = 0;

mod errno;
mod fs;
mod mm;
mod process;
mod signal;

pub use fs::*;
pub use mm::*;
pub use process::*;
pub use signal::*;

use crate::fs::IOVec;
use crate::sbi::shutdown;
use crate::signal::{SignalAction, SignalFlags};
use crate::timer::{Timespec, Timeval, Tms};
use crate::types::FdSet;
use crate::utsname::*;

pub fn syscall(syscall_id: usize, args: [usize; 6]) -> isize {
    // if syscall_id != SYSCALL_READ && syscall_id != SYSCALL_WRITE {
    //     println!(
    //         "\x1b[032m[kernel]\x1b[0m syscall: id = {}, args = {:x?}",
    //         syscall_id, args
    //     );
    // }
    match syscall_id {
        SYSCALL_GETCWD => sys_getcwd(args[0] as *mut u8, args[1] as usize),
        SYSCALL_DUP => sys_dup(args[0]),
        SYSCALL_DUP3 => sys_dup3(args[0] as usize, args[1] as usize),
        SYSCALL_FCNTL => sys_fcntl(args[0] as usize, args[1] as usize, args[2] as usize),
        SYSCALL_IOCTL => sys_ioctl(args[0] as usize, args[1] as usize, args[2] as usize),
        SYSCALL_MKDIRAT => sys_mkdirat(args[0] as isize, args[1] as *const u8, args[2] as u32),
        SYSCALL_UNLINKAT => sys_unlinkat(args[0] as isize, args[1] as *const u8, args[2] as u32),
        SYSCALL_UMOUNT2 => sys_umount(args[0] as *const u8, args[1] as usize),
        SYSCALL_MOUNT => sys_mount(
            args[0] as *const u8,
            args[1] as *const u8,
            args[2] as *const u8,
            args[3] as usize,
            args[4] as *const u8,
        ),
        SYSCALL_FACCESSAT => sys_faccessat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as u32,
            args[3] as usize,
        ),
        SYSCALL_CHDIR => sys_chdir(args[0] as *const u8),
        SYSCALL_OPENAT => sys_openat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as u32,
            args[3] as u32,
        ),
        SYSCALL_CLOSE => sys_close(args[0]),
        SYSCALL_PIPE => sys_pipe2(args[0] as *mut u32),
        SYSCALL_GETDENTS64 => {
            sys_getdents64(args[0] as usize, args[1] as *const u8, args[2] as usize)
        }
        SYSCALL_LSEEK => sys_lseek(args[0] as usize, args[1] as usize, args[2] as usize),
        SYSCALL_READ => sys_read(args[0], args[1] as *const u8, args[2]),
        SYSCALL_WRITE => sys_write(args[0], args[1] as *const u8, args[2]),
        SYSCALL_READV => sys_readv(args[0], args[1] as *const IOVec, args[2]),
        SYSCALL_WRITEV => sys_writev(args[0], args[1] as *const IOVec, args[2]),
        SYSCALL_SENDFILE => sys_sendfile(
            args[0] as usize,
            args[1] as usize,
            args[2] as *mut usize,
            args[3] as usize,
        ),
        SYSCALL_PSELECT6 => sys_pselect6(
            args[0] as usize,
            args[1] as *mut FdSet,
            args[2] as *mut FdSet,
            args[3] as *mut FdSet,
            args[4] as *const Timespec,
            args[5] as usize,
        ),
        SYSCALL_READLINKAT => sys_readlinkat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as *const u8,
            args[3] as u32,
        ),
        SYSCALL_FSTATAT => sys_fstatat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as *const u8,
            args[3] as usize,
        ),
        SYSCALL_FSTAT => sys_fstat(args[0] as usize, args[1] as *const u8),
        SYSCALL_UTIMENSAT => sys_utimensat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as *const Timespec,
            args[3] as usize,
        ),
        SYSCALL_EXIT => sys_exit(args[0] as i32),
        SYSCALL_EXIT_GROUP => sys_exit_group(args[0] as i32),
        SYSCALL_SET_TID_ADDRESS => sys_set_tid_address(args[0] as usize),
        SYSCALL_NANOSLEEP => sys_nanosleep(args[0] as *mut Timespec, args[1] as *mut Timespec),
        SYSCALL_CLOCK_GETTIME => sys_clock_gettime(args[0] as usize, args[1] as *mut Timespec),
        SYSCALL_SYSLOG => 0,
        SYSCALL_SCHED_YIELD => sys_sched_yield(),
        SYSCALL_KILL => sys_kill(args[0] as isize, args[1] as u64),
        SYSCALL_RT_SIGACTION => sys_rt_sigaction(
            args[0] as usize,
            args[1] as *const SignalAction,
            args[2] as *mut SignalAction,
            args[3] as usize,
        ),
        SYSCALL_RT_SIGPROCMASK => sys_rt_sigprocmask(
            args[0] as u32,
            args[1] as *const SignalFlags,
            args[2] as *mut SignalFlags,
            args[3] as usize,
        ),
        SYSCALL_RT_SIGTIMEDWAIT => sys_rt_sigtimedwait(
            args[0] as *const SignalFlags,
            args[1] as usize,
            args[2] as *const Timespec,
            args[3] as usize,
        ),
        SYSCALL_RT_SIGRETURN => sys_rt_sigretrun(),
        SYSCALL_TIMES => sys_times(args[0] as *mut Tms),
        SYSCALL_UNAME => sys_uname(args[0] as *mut Utsname),
        SYSCALL_GETRUSAGE => 0,
        SYSCALL_UMASK => 0,
        SYSCALL_GET_TIME_OF_DAY => sys_get_time_of_day(args[0] as *mut Timeval),
        SYSCALL_BRK => sys_brk(args[0]),
        SYSCALL_GETPID => sys_getpid(),
        SYSCALL_GETPPID => sys_getppid(),
        SYSCALL_GETUID => sys_getuid(),
        SYSCALL_GETTID => sys_gettid(),
        SYSCALL_SYSINFO => 0,
        SYSCALL_CLONE => sys_clone(
            args[0] as u32,
            args[1] as usize,
            args[2] as *mut u32,
            args[3] as usize,
            args[4] as *mut u32,
        ),
        SYSCALL_EXECVE => sys_execve(
            args[0] as *const u8,
            args[1] as *const usize,
            args[2] as *const usize,
        ),
        SYSCALL_WAIT4 => sys_wait4(
            args[0] as isize,
            args[1] as *mut i32,
            args[2] as u32,
            args[3] as usize,
        ),
        SYSCALL_MMAP => sys_mmap(
            args[0] as usize,
            args[1] as usize,
            args[2] as u32,
            args[3] as u32,
            args[4] as usize,
            args[5] as usize,
        ),
        SYSCALL_MPROTECT => sys_mprotect(args[0] as usize, args[1] as usize, args[2] as u32),
        SYSCALL_MUNMAP => sys_munmap(args[0] as usize, args[1] as usize),
        SYSCALL_PRLIMIT => 0,
        SYSCALL_RENAMEAT2 => sys_renameat2(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as isize,
            args[3] as *const u8,
            args[4] as u32,
        ),
        SYSCALL_SHUTDOWN => shutdown(),
        _ => {
            println!("Unsupported syscall_id: {}", syscall_id);
            sys_exit(0)
        }
    }
}
