use crate::config::*;
use crate::mm::{MapPermission, VirtAddr, KERNEL_SPACE};
use crate::task::TaskControlBlock;
use alloc::collections::BTreeMap;
use alloc::sync::Arc;
use alloc::vec;
use alloc::vec::Vec;
use lazy_static::*;
use spin::Mutex;

use super::INITPROC;

pub struct RecycleAllocator {
    current: usize,
    recycled: Vec<usize>,
}

impl RecycleAllocator {
    pub fn new() -> Self {
        RecycleAllocator {
            current: 0,
            recycled: Vec::new(),
        }
    }
    pub fn alloc(&mut self) -> usize {
        if let Some(id) = self.recycled.pop() {
            id
        } else {
            self.current += 1;
            self.current - 1
        }
    }
    pub fn dealloc(&mut self, id: usize) {
        assert!(id < self.current);
        assert!(
            !self.recycled.iter().any(|i| *i == id),
            "id {} has been deallocated!",
            id
        );
        self.recycled.push(id);
    }
}

lazy_static! {
    static ref PID_ALLOCATOR: Mutex<RecycleAllocator> = Mutex::new(RecycleAllocator::new());
    static ref KSTACK_ALLOCATOR: Mutex<RecycleAllocator> = Mutex::new(RecycleAllocator::new());
}

pub struct PidHandle(pub usize);

pub fn pid_alloc() -> PidHandle {
    PidHandle(PID_ALLOCATOR.lock().alloc())
}

impl Drop for PidHandle {
    fn drop(&mut self) {
        PID_ALLOCATOR.lock().dealloc(self.0);
    }
}

/// Return (bottom, top) of a kernel stack in kernel space.
pub fn kernel_stack_position(kstack_id: usize) -> (usize, usize) {
    let top = TRAMPOLINE - kstack_id * (KERNEL_STACK_SIZE + PAGE_SIZE);
    let bottom = top - KERNEL_STACK_SIZE;
    (bottom, top)
}

pub struct KernelStack(pub usize);

pub fn kstack_alloc() -> KernelStack {
    let kstack_id = KSTACK_ALLOCATOR.lock().alloc();
    let (kstack_bottom, kstack_top) = kernel_stack_position(kstack_id);
    KERNEL_SPACE.insert_framed_area(
        kstack_bottom.into(),
        kstack_top.into(),
        MapPermission::R | MapPermission::W,
    );
    KernelStack(kstack_id)
}

impl Drop for KernelStack {
    fn drop(&mut self) {
        let (kernel_stack_bottom, _) = kernel_stack_position(self.0);
        let kernel_stack_bottom_va: VirtAddr = kernel_stack_bottom.into();
        KERNEL_SPACE.remove_area_with_start_vpn(kernel_stack_bottom_va.into());
    }
}

impl KernelStack {
    #[allow(unused)]
    pub fn push_on_top<T>(&self, value: T) -> *mut T
    where
        T: Sized,
    {
        let kernel_stack_top = self.get_top();
        let ptr_mut = (kernel_stack_top - core::mem::size_of::<T>()) as *mut T;
        unsafe {
            *ptr_mut = value;
        }
        ptr_mut
    }
    pub fn get_top(&self) -> usize {
        let (_, kernel_stack_top) = kernel_stack_position(self.0);
        kernel_stack_top
    }
}
/// - `tid` -> `TCB`
pub const PIDTYPE_TID: usize = 0;
/// - `pid` -> 线程组
pub const PIDTYPE_TGID: usize = 1;
/// - `ppid` -> 子进程
pub const PIDTYPE_PPID: usize = 2;

lazy_static! {
    pub static ref PID_HASH: Vec<Mutex<BTreeMap<usize, Vec<Arc<TaskControlBlock>>>>> = {
        let mut res = Vec::with_capacity(3);
        for _ in 0..3 {
            res.push(Mutex::new(BTreeMap::new()));
        }
        res
    };
}

pub fn insert_into_tid2tcb(tid: usize, task: Arc<TaskControlBlock>) {
    if PID_HASH[PIDTYPE_TID]
        .lock()
        .insert(tid, vec![task])
        .is_some()
    {
        println!("tid {:?} has been inserted into hash table", tid);
    }
}

pub fn insert_into_thread_group(pid: usize, task: Arc<TaskControlBlock>) {
    let mut hash = PID_HASH[PIDTYPE_TGID].lock();
    if let Some(vec) = hash.get_mut(&pid) {
        vec.push(task);
    } else {
        hash.insert(pid, vec![task]);
    }
}

pub fn insert_child_process(ppid: usize, task: Arc<TaskControlBlock>) {
    let mut hash = PID_HASH[PIDTYPE_PPID].lock();
    if let Some(vec) = hash.get_mut(&ppid) {
        vec.push(task);
    } else {
        hash.insert(ppid, vec![task]);
    }
}

/// 从进程散列表中移除线程号为 `tid` 的进程
pub fn remove_from_tid2tcb(tid: usize) {
    PID_HASH[PIDTYPE_TID].lock().remove(&tid);
}

/// 从线程组散列表中移除线程组号为 `pid` 的线程组
///
/// 该函数会同时删除进程散列表中的相应进程
pub fn remove_from_thread_group(pid: usize) {
    let threads = PID_HASH[PIDTYPE_TGID].lock().remove(&pid).unwrap();
    for thread in threads.iter() {
        remove_from_tid2tcb(thread.get_tid());
    }
}

/// 将线程组号为 `ppid` 的线程组的所有子线程组移到 `init` 进程下
pub fn move_child_process_to_init(ppid: usize) {
    let mut hash = PID_HASH[PIDTYPE_PPID].lock();
    if let Some(vec) = hash.remove(&ppid) {
        let init_children = hash.get_mut(&INITPROC.get_pid()).unwrap();
        for child in vec {
            child.inner_exclusive_access().parent = Some(Arc::downgrade(&INITPROC));
            init_children.push(child);
        }
    }
}
