use super::{
    add_task, hartid, insert_child_process, insert_into_thread_group, insert_into_tid2tcb,
    kstack_alloc, pid_alloc, Aux, KernelStack, PidHandle, TaskContext, AT_EXECFN, AT_NULL,
    AT_RANDOM,
};
use crate::config::*;
use crate::fs::{root_inode, FileClass, Files, OSInode, Stdin, Stdout};
use crate::mm::{
    kernel_token, translated_refmut, MapPermission, MemorySet, PhysPageNum, VirtAddr, KERNEL_SPACE,
};
use crate::signal::{SignalActions, SignalFlags, SignalInfo};
use crate::syscall::CloneFlags;
use crate::trap::{trap_handler, TrapContext};
use alloc::string::String;
use alloc::sync::{Arc, Weak};
use alloc::vec;
use alloc::vec::Vec;
use spin::{Mutex, MutexGuard};

pub struct TaskControlBlock {
    // 不可变的
    tid: PidHandle,
    pub pid: usize,
    pub ppid: usize,
    pub kstack: KernelStack,

    // 可变的
    inner: Mutex<TaskControlBlockInner>,
}

impl TaskControlBlock {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, TaskControlBlockInner> {
        self.inner.lock()
    }

    pub fn get_tid(&self) -> usize {
        self.tid.0
    }

    pub fn get_pid(&self) -> usize {
        self.pid
    }

    pub fn get_ppid(&self) -> usize {
        self.ppid
    }
}

pub struct TaskControlBlockInner {
    /// 用户栈虚拟地址
    ///
    /// 用户可能自己指定栈顶，此时的用户栈资源将不由内核控制
    pub ustack_top: Option<VirtAddr>,
    /// trap 上下文虚拟地址
    pub trap_cx_bottom: VirtAddr,
    /// trap 上下文所在物理页号
    pub trap_cx_ppn: PhysPageNum,
    /// 用户地址空间
    pub memory_set: Arc<MemorySet>,
    /// 当前工作目录
    pub cwd: Arc<OSInode>,
    /// 打开文件表
    pub fd_table: Arc<Files>,
    pub parent: Option<Weak<TaskControlBlock>>,
    pub set_child_tid: Option<VirtAddr>,
    pub clear_child_tid: Option<VirtAddr>,
    pub task_cx: TaskContext,
    pub task_status: TaskStatus,
    pub exit_code: Option<i32>,
    /// 信号描述符
    pub signal_info: Arc<SignalInfo>,
    /// 共享的信号处理程序
    pub signal_actions: Arc<SignalActions>,
    /// 信号掩码
    pub signal_mask: SignalFlags,
    /// 挂起的信号
    pub signal_pending: SignalFlags,
}

impl TaskControlBlockInner {
    pub fn get_user_token(&self) -> usize {
        self.memory_set.token()
    }

    pub fn alloc_fd(&self) -> usize {
        self.fd_table.alloc_fd()
    }

    pub fn get_trap_cx(&self) -> &'static mut TrapContext {
        self.trap_cx_ppn.get_mut()
    }

    /// 分配用户资源（用户栈和 `Trap` 上下文）
    pub fn alloc_user_res(&mut self) {
        // 分配用户栈
        let (_, ustack_top) = self.memory_set.insert_framed_area_randomly(
            USER_STACK_SIZE,
            Some(VirtAddr::from(USTACK_TOP)),
            MapPermission::R | MapPermission::W | MapPermission::U,
        );
        // 分配 Trap 上下文
        let (trap_cx_bottom, _) = self.memory_set.insert_framed_area_randomly(
            PAGE_SIZE,
            Some(VirtAddr::from(TRAP_CONTEXT_TOP)),
            MapPermission::R | MapPermission::W,
        );
        let trap_cx_ppn = self
            .memory_set
            .translate(trap_cx_bottom.into())
            .unwrap()
            .ppn();
        self.ustack_top = Some(ustack_top);
        self.trap_cx_bottom = trap_cx_bottom;
        self.trap_cx_ppn = trap_cx_ppn;
    }

    /// 复制用户资源（用户栈和 `Trap` 上下文）
    pub fn clone_user_res(&mut self, another: &TaskControlBlockInner) {
        self.alloc_user_res();

        // 复制用户栈
        self.memory_set.clone_area(
            (self.ustack_top.unwrap() - USER_STACK_SIZE).into(),
            &another.memory_set,
        );

        // 复制 Trap 上下文
        self.memory_set
            .clone_area(self.trap_cx_bottom.into(), &another.memory_set);
    }

    /// 释放用户资源（用户栈和 `Trap` 上下文）
    pub fn dealloc_user_res(&self) {
        if let Some(ustack_top) = self.ustack_top {
            self.memory_set
                .remove_area_with_start_vpn((ustack_top - USER_STACK_SIZE).into());
        }
        self.memory_set
            .remove_area_with_start_vpn(self.trap_cx_bottom.into());
    }
}

impl TaskControlBlock {
    /// 只有 `init` 进程创建的时候调用该函数
    pub fn new(elf_data: &[u8]) -> Arc<Self> {
        // memory_set with elf program headers/trampoline
        let (memory_set, entry_point, _) = MemorySet::from_elf(elf_data);
        // alloc a pid/tid and a kernel stack in kernel space
        let tid = pid_alloc();
        let pid = tid.0;
        let kstack = kstack_alloc();
        let kstack_top = kstack.get_top();
        let task = Arc::new(Self {
            tid,
            pid,
            ppid: 0,
            kstack,
            inner: Mutex::new(TaskControlBlockInner {
                ustack_top: None,
                trap_cx_bottom: 0.into(),
                trap_cx_ppn: 0.into(),
                memory_set: Arc::new(memory_set),
                cwd: root_inode(),
                fd_table: Arc::new(Files::new(vec![
                    // 0 -> stdin
                    Some(FileClass::Other(Arc::new(Stdin))),
                    // 1 -> stdout
                    Some(FileClass::Other(Arc::new(Stdout))),
                    // 2 -> stderr
                    Some(FileClass::Other(Arc::new(Stdout))),
                ])),
                signal_info: Arc::new(SignalInfo::new()),
                signal_actions: Arc::new(SignalActions::default()),
                parent: None,
                task_cx: TaskContext::goto_trap_return(kstack_top),
                task_status: TaskStatus::Ready,
                exit_code: None,
                signal_mask: SignalFlags::empty(),
                signal_pending: SignalFlags::empty(),
                set_child_tid: None,
                clear_child_tid: None,
            }),
        });

        let mut task_inner = task.inner_exclusive_access();
        task_inner.alloc_user_res();
        let trap_cx = task_inner.get_trap_cx();
        let ustack_top = task_inner.ustack_top.unwrap().into();
        let kstack_top = task.kstack.get_top();
        *trap_cx = TrapContext::app_init_context(
            entry_point,
            ustack_top,
            kernel_token(),
            kstack_top,
            trap_handler as usize,
            hartid(),
        );
        drop(task_inner);

        insert_into_tid2tcb(task.get_tid(), Arc::clone(&task));
        insert_into_thread_group(task.get_pid(), Arc::clone(&task));
        add_task(Arc::clone(&task));
        task
    }

    /// 未考虑多线程情况
    pub fn exec(&self, elf_data: &[u8], args: Vec<String>, envs: Vec<String>) {
        // memory_set with elf program headers/trampoline
        let (memory_set, entry_point, mut auxv) = MemorySet::from_elf(elf_data);
        let token = memory_set.token();
        // 替换用户地址空间
        let mut task_inner = self.inner_exclusive_access();
        // CLONE_CHILD_CLEARTID
        if let Some(ptr) = task_inner.clear_child_tid {
            let pre_token = task_inner.get_user_token();
            *translated_refmut(pre_token, usize::from(ptr) as *mut u32) = 0;
        }
        task_inner.memory_set = Arc::new(memory_set);
        // 重新分配用户资源
        task_inner.alloc_user_res();

        // 栈上参数放置
        // -------------------------------------------------------------------
        let mut user_sp = task_inner.ustack_top.unwrap().into();
        let mut argv = Vec::new();
        let mut envp = Vec::new();

        // env_str
        for i in 0..envs.len() {
            user_sp -= envs[i].len() + 1;
            envp.push(user_sp);
            let mut p = user_sp;
            for c in envs[i].as_bytes() {
                *translated_refmut(token, p as *mut u8) = *c;
                p += 1;
            }
            *translated_refmut(token, p as *mut u8) = 0;
        }
        envp.push(0);
        user_sp -= user_sp % core::mem::size_of::<usize>();

        // arg_str
        for i in 0..args.len() {
            user_sp -= args[i].len() + 1;
            argv.push(user_sp);
            let mut p = user_sp;
            for c in args[i].as_bytes() {
                *translated_refmut(token, p as *mut u8) = *c;
                p += 1;
            }
            *translated_refmut(token, p as *mut u8) = 0;
        }
        argv.push(0);
        // make the user_sp aligned to 8B for k210 platform
        user_sp -= user_sp % core::mem::size_of::<usize>();

        // aux
        user_sp -= 16;
        let mut p = user_sp;
        for i in 0..0xf {
            *translated_refmut(token, p as *mut u8) = i as u8;
            p += 1;
        }
        auxv.push(Aux::new(AT_RANDOM, user_sp));
        auxv.push(Aux::new(AT_EXECFN, argv[0]));
        auxv.push(Aux::new(AT_NULL, 0));
        user_sp -= auxv.len() * core::mem::size_of::<Aux>();
        let auxv_base = user_sp;
        for aux in 0..auxv.len() {
            *translated_refmut(
                token,
                (auxv_base + aux * core::mem::size_of::<Aux>()) as *mut Aux,
            ) = auxv[aux];
        }

        // envp
        user_sp -= envp.len() * core::mem::size_of::<usize>();
        let envp_base = user_sp;
        for env in 0..envp.len() {
            *translated_refmut(
                token,
                (envp_base + env * core::mem::size_of::<usize>()) as *mut usize,
            ) = envp[env];
        }

        // argv
        user_sp -= argv.len() * core::mem::size_of::<usize>();
        let argv_base = user_sp;
        for arg in 0..argv.len() {
            *translated_refmut(
                token,
                (argv_base + arg * core::mem::size_of::<usize>()) as *mut usize,
            ) = argv[arg];
            // println!("argv[{}] = {:#x}", arg, argv[arg]);
        }

        // argc
        user_sp -= core::mem::size_of::<usize>();
        *translated_refmut(token, user_sp as *mut usize) = args.len();
        // -------------------------------------------------------------------

        // println!("exec: user sp = {:#x}", user_sp);
        // println!("in exec: argv = {:#x}, envp = {:#x}, auxv = {:#x}", argv_base, envp_base, auxv_base);

        // initialize trap_cx
        let mut trap_cx = TrapContext::app_init_context(
            entry_point,
            user_sp,
            KERNEL_SPACE.token(),
            self.kstack.get_top(),
            trap_handler as usize,
            hartid(),
        );
        trap_cx.x[10] = args.len();
        trap_cx.x[11] = argv_base;
        *task_inner.get_trap_cx() = trap_cx;
    }

    /// 创建轻量级进程
    pub fn clone(
        self: &Arc<Self>,
        flags: CloneFlags,
        stack: Option<VirtAddr>,
        parent_tid: *mut u32,
        _tls: usize,
        child_tid: *mut u32,
    ) -> Arc<Self> {
        let parent_inner = self.inner_exclusive_access();
        // 分配 tid
        let tid = pid_alloc();
        // 分配内核栈
        let kstack = kstack_alloc();
        let kstack_top = kstack.get_top();
        // 检查是否共享虚拟内存
        let memory_set = if flags.contains(CloneFlags::CLONE_VM) {
            Arc::clone(&parent_inner.memory_set)
        } else {
            Arc::new(MemorySet::from_existed_user(&parent_inner.memory_set))
        };
        // 检查是否共享打开文件表
        let fd_table = if flags.contains(CloneFlags::CLONE_FILES) {
            Arc::clone(&parent_inner.fd_table)
        } else {
            Arc::new(Files::clone(&parent_inner.fd_table))
        };
        // 检查是否共享信号处理程序表
        let signal_actions = if flags.contains(CloneFlags::CLONE_SIGHAND) {
            Arc::clone(&parent_inner.signal_actions)
        } else {
            Arc::new(SignalActions::clone(&parent_inner.signal_actions))
        };
        // CLONE_PARENT_SETTID
        if flags.contains(CloneFlags::CLONE_PARENT_SETTID) {
            let parent_token = parent_inner.get_user_token();
            *translated_refmut(parent_token, parent_tid) = tid.0 as u32;
        }
        // set_child_tid, clear_child_tid
        let set_child_tid = if flags.contains(CloneFlags::CLONE_CHILD_SETTID) {
            Some(VirtAddr::from(child_tid as usize))
        } else {
            None
        };
        let clear_child_tid = if flags.contains(CloneFlags::CLONE_CHILD_CLEARTID) {
            Some(VirtAddr::from(child_tid as usize))
        } else {
            None
        };
        // 检查是否创建线程
        let (pid, ppid, parent, signal_info);
        if flags.contains(CloneFlags::CLONE_THREAD) {
            pid = self.pid;
            ppid = self.ppid;
            parent = Option::clone(&parent_inner.parent);
            signal_info = Arc::clone(&parent_inner.signal_info);
        } else {
            pid = tid.0;
            ppid = self.pid;
            parent = Some(Arc::downgrade(self));
            signal_info = Arc::new(SignalInfo::new());
        }

        let child = Arc::new(TaskControlBlock {
            tid,
            pid,
            ppid,
            kstack,
            inner: Mutex::new(TaskControlBlockInner {
                ustack_top: None,
                trap_cx_bottom: 0.into(),
                trap_cx_ppn: 0.into(),
                memory_set,
                cwd: Arc::clone(&parent_inner.cwd),
                fd_table,
                signal_info,
                signal_actions,
                parent,
                task_cx: TaskContext::goto_trap_return(kstack_top),
                task_status: TaskStatus::Ready,
                exit_code: None,
                signal_mask: parent_inner.signal_mask,
                signal_pending: parent_inner.signal_pending,
                set_child_tid,
                clear_child_tid,
            }),
        });

        let mut child_inner = child.inner_exclusive_access();
        // 分配用户资源
        if flags.contains(CloneFlags::CLONE_THREAD) {
            child_inner.alloc_user_res();
        } else {
            child_inner.clone_user_res(&parent_inner);
        }
        let trap_cx = child_inner.get_trap_cx();
        trap_cx.kernel_sp = kstack_top;
        if let Some(sp) = stack {
            let ustack_top = child_inner.ustack_top.take().unwrap();
            child_inner
                .memory_set
                .remove_area_with_start_vpn((ustack_top - USER_STACK_SIZE).into());
            // child_inner.ustack_top = None;
            trap_cx.set_sp(sp.into());
        }
        // CLONE_CHILD_SETTID
        if flags.contains(CloneFlags::CLONE_CHILD_SETTID) {
            let child_token = child_inner.get_user_token();
            *translated_refmut(child_token, child_tid) = child.get_tid() as u32;
        }
        drop(child_inner);

        insert_into_tid2tcb(child.get_tid(), Arc::clone(&child));
        insert_into_thread_group(child.get_pid(), Arc::clone(&child));
        if !flags.contains(CloneFlags::CLONE_THREAD) {
            // 子进程的散列表只保留子线程组的主线程
            insert_child_process(child.get_ppid(), Arc::clone(&child));
        }
        add_task(Arc::clone(&child));

        child
    }
}

#[derive(Copy, Clone, PartialEq)]
pub enum TaskStatus {
    Ready,
    Running,
    Zombie,
    Stopped,
}
