#![no_std]
#![no_main]
#![allow(clippy::println_empty_string)]

#[macro_use]
extern crate alloc;

#[macro_use]
extern crate user_lib;

const LF: u8 = 0x0au8;
const CR: u8 = 0x0du8;
const DL: u8 = 0x7fu8;
const BS: u8 = 0x08u8;
const LINE_START: &str = ">> ";

use alloc::string::{String, ToString};
use alloc::vec::Vec;
use user_lib::console::getchar;
use user_lib::{close, dup, exec, exit, fork, open, pipe, shutdown, waitpid, OpenFlags};

#[derive(Debug)]
struct ProcessArguments {
    input: String,
    output: String,
    flags: OpenFlags,
    args_copy: Vec<String>,
    args_addr: Vec<*const u8>,
}

impl ProcessArguments {
    pub fn new(command: &str) -> Self {
        let mut args_copy: Vec<String> = Vec::new();
        if command.starts_with("busybox echo ") {
            args_copy.push(String::from("busybox\0"));
            args_copy.push(String::from("echo\0"));
            let mut echo_arg = String::new();
            let mut args = Vec::new();
            for (id, c) in command.chars().enumerate() {
                if id < 13 {
                    continue;
                }
                match c {
                    '\"' => continue,
                    '>' | '<' => {
                        echo_arg = echo_arg.trim_end().to_string();
                        args = command[id..]
                            .split(" ")
                            .map(|tmp| String::from(tmp) + "\0")
                            .collect::<Vec<_>>();
                        break;
                    }
                    _ => echo_arg.push(c),
                }
            }
            echo_arg.push('\0');
            args_copy.push(echo_arg);
            if !args.is_empty() {
                args_copy.extend(args);
            }
        } else if command.starts_with("busybox printf ") {
            args_copy.push(String::from("busybox\0"));
            args_copy.push(String::from("printf\0"));
            let mut printf_arg = String::new();
            let left = command.find('\"').unwrap();
            let right = command.rfind('\"').unwrap();
            printf_arg.push_str(&command[left + 1..right]);
            printf_arg.push('\0');
            args_copy.push(printf_arg);
        } else if command.starts_with("busybox which ") {
            args_copy.push(String::from("busybox\0"));
            args_copy.push(String::from("which\0"));
            args_copy.push(String::from("busybox\0"));
        } else {
            let args: Vec<_> = command.split(' ').collect();
            args_copy = args
                .iter()
                .filter(|&arg| !arg.is_empty())
                .map(|&arg| {
                    let mut string = String::new();
                    string.push_str(arg);
                    string.push('\0');
                    string
                })
                .collect();
        }
        // println!("args_copy = {:?}", args_copy);

        // redirect input
        let mut input = String::new();
        if let Some((idx, _)) = args_copy
            .iter()
            .enumerate()
            .find(|(_, arg)| arg.as_str() == "<\0")
        {
            input = args_copy[idx + 1].clone();
            args_copy.drain(idx..=idx + 1);
        }

        // redirect output
        let mut output = String::new();
        let mut flags = OpenFlags::empty();
        if let Some((idx, arg)) = args_copy
            .iter()
            .enumerate()
            .find(|(_, arg)| arg.as_str() == ">\0" || arg.as_str() == ">>\0")
        {
            if arg.as_str() == ">\0" {
                flags |= OpenFlags::CREATE;
            } else if arg.as_str() == ">>\0" {
                flags |= OpenFlags::O_APPEND;
            }
            output = args_copy[idx + 1].clone();
            args_copy.drain(idx..=idx + 1);
        }

        let mut args_addr: Vec<*const u8> = args_copy.iter().map(|arg| arg.as_ptr()).collect();
        args_addr.push(core::ptr::null::<u8>());

        Self {
            input,
            output,
            flags,
            args_copy,
            args_addr,
        }
    }
}

#[allow(unused)]
fn autotest() {
    println!("------------ autotest -------------");
    let mut test = Vec::new();
    test.push("brk");
    test.push("chdir");
    test.push("clone");
    test.push("close");
    test.push("dup");
    test.push("dup2");
    test.push("execve");
    test.push("exit");
    test.push("fork");
    test.push("fstat");
    test.push("getcwd");
    test.push("getdents");
    test.push("getpid");
    test.push("getppid");
    test.push("gettimeofday");
    test.push("mkdir_");
    test.push("mmap");
    test.push("mount");
    test.push("munmap");
    test.push("open");
    test.push("openat");
    test.push("pipe");
    test.push("read");
    test.push("sleep");
    test.push("times");
    test.push("umount");
    test.push("uname");
    test.push("unlink");
    test.push("wait");
    test.push("waitpid");
    test.push("write");
    test.push("yield");

    for cmd in test.iter() {
        parse_cmd_and_exec(cmd.to_string());
    }
    println!("------------ autotest finish -------------");
}

#[allow(unused)]
fn libc_test() {
    println!("------------ libc_test -------------");
    let mut cmds = vec![
        "argv",
        "basename",
        "clocale_mbfuncs",
        "clock_gettime",
        "crypt",
        "dirname",
        "env",
        // "fdopen",
        "fnmatch",
        // "fscanf",
        // "fwscanf",
        "iconv_open",
        "inet_pton",
        "mbc",
        "memstream",
        "pthread_cancel_points",
        "pthread_cancel",
        "pthread_cond",
        "pthread_tsd",
        // "qsort",
        // "random",
        // "search_hsearch",
        // "search_insque",
        // "search_lsearch",
        // "search_tsearch",
        // // "setjmp",
        // "snprintf",
        // // "socket",
        // "sscanf",
        // // "sscanf_long",
        // // "stat",
        // "strftime",
        // "string",
        // "string_memcpy",
        // "string_memmem",
        // "string_memset",
        // "string_strchr",
        // "string_strcspn",
        // "string_strstr",
        // // "strptime",
        // "strtod",
        // "strtod_simple",
        // "strtof",
        // "strtol",
        // "strtold",
        // "swprintf",
        // "tgmath",
        // "time",
        // "tls_align",
        // "udiv",
        // // "ungetc",
        // // "utime",
        // "wcsstr",
        // "wcstol",
        // "pleval",
        // // "daemon_failure",
        // "dn_expand_empty",
        // "dn_expand_ptr_0",
        // // "fflush_exit",
        // "fgets_eof",
        // // "fgetwc_buffering",
        // "fpclassify_invalid_ld80",
        // // "ftello_unflushed_append",
        // "getpwnam_r_crash",
        // "getpwnam_r_errno",
        // "iconv_roundtrips",
        // "inet_ntop_v4mapped",
        // "inet_pton_empty_last_field",
        // "iswspace_null",
        // "lrand48_signextend",
        // // "lseek_large",
        // "malloc_0",
        // "mbsrtowcs_overflow",
        // "memmem_oob_read",
        // "memmem_oob",
        // "mkdtemp_failure",
        // "mkstemp_failure",
        // "printf_1e9_oob",
        // "printf_fmt_g_round",
        // "printf_fmt_g_zeros",
        // "printf_fmt_n",
        // // "pthread_robust_detach",
        // // "pthread_cancel_sem_wait",
        // // "pthread_cond_smasher",
        // // "pthread_condattr_setclock",
        // // "pthread_exit_cancel",
        // // "pthread_once_deadlock",
        // // "pthread_rwlock_ebusy",
        // // "putenv_doublefree",
        // "regex_backref_0",
        // "regex_bracket_icase",
        // "regex_ere_backref",
        // "regex_escaped_high_byte",
        // "regex_negated_range",
        // "regexec_nosub",
        // // "rewind_clear_error",
        // // "rlimit_open_files",
        // "scanf_bytes_consumed",
        // "scanf_match_literal_eof",
        // "scanf_nullbyte_char",
        // "setvbuf_unget",
        // // "sigprocmask_internal",
        // "sscanf_eof",
        // // "statvfs",
        // "strverscmp",
        // // "syscall_sign_extend",
        // "uselocale_0",
        // "wcsncpy_read_overflow",
        // "wcsstr_false_negative",
    ];

    for cmd in cmds.iter() {
        parse_cmd_and_exec(String::from("runtest.exe -w entry-static.exe ") + cmd);
    }
    println!("------------ libc_test finish -------------");
}

#[allow(unused)]
fn busybox_test() {
    println!("------------ busybox_test -------------");
    let mut cmds = vec![
        "echo \"#### independent command test\"",
        "ash -c exit",
        "sh -c exit",
        "basename /aaa/bbb",
        "cal",
        "clear",
        "date",
        // "df",
        "dirname /aaa/bbb",
        "dmesg",
        "du",
        "expr 1 + 1",
        "false",
        "true",
        "which ls",
        "uname",
        "uptime",
        "printf \"abc\\n\"",
        // "ps",
        "pwd",
        // "free",
        // "hwclock",
        // "kill 10",
        "ls",
        "sleep 1",
        "echo \"#### file opration test\"",
        "touch test.txt",
        "echo \"hello world\" > test.txt",
        "cat test.txt",
        "cut -c 3 test.txt",
        "od test.txt",
        "head test.txt",
        "tail test.txt",
        "hexdump -C test.txt",
        "md5sum test.txt",
        "echo \"ccccccc\" >> test.txt",
        "echo \"bbbbbbb\" >> test.txt",
        "echo \"aaaaaaa\" >> test.txt",
        "echo \"2222222\" >> test.txt",
        "echo \"1111111\" >> test.txt",
        "echo \"bbbbbbb\" >> test.txt",
        "sort test.txt | ./busybox uniq",
        "stat test.txt",
        "strings test.txt",
        "wc test.txt",
        "[ -f test.txt ]",
        "more test.txt",
        "rm test.txt",
        "mkdir test_dir",
        "mv test_dir test",
        "rmdir test",
        "grep hello busybox_cmd.txt",
        "cp busybox_cmd.txt busybox_cmd.bak",
        "rm busybox_cmd.bak",
        "find -name \"busybox_cmd.txt\"",
    ];

    for cmd in cmds.iter() {
        println!("------------ busybox_test: {} -------------", cmd);
        if parse_cmd_and_exec(String::from("busybox ") + cmd) != 0 && cmd != &"false" {
            println!("testcase busybox {} fail", cmd);
        } else {
            println!("testcase busybox {} success", cmd);
        }
        println!("---------------------------------------\n");
    }
    println!("------------ busybox_test finish -------------");
}

#[allow(unused)]
fn lua_test() {
    println!("------------ lua_test -------------");
    let mut cmds = vec![
        "date.lua",
        "file_io.lua",
        "max_min.lua",
        "random.lua",
        "remove.lua",
        "round_num.lua",
        "sin30.lua",
        "sort.lua",
        "strings.lua",
    ];

    for cmd in cmds.iter() {
        println!("------------ lua_test: {} -------------", cmd);
        if parse_cmd_and_exec(String::from("lua ") + cmd) != 0 {
            println!("testcase lua {} fail", cmd);
        } else {
            println!("testcase lua {} success", cmd);
        }
        println!("---------------------------------------\n");
    }
    println!("------------ lua_test finish -------------");
}

#[allow(unused)]
fn lmbench_test() {
    println!("------------ lmbench_test -------------");
    let mut cmds = vec![
        "busybox echo latency measurements",
        "lmbench_all lat_syscall -P 1 null",
        "lmbench_all lat_syscall -P 1 read",
        "lmbench_all lat_syscall -P 1 write",
        "busybox mkdir -p /var/tmp",
        "busybox touch /var/tmp/lmbench",
        "lmbench_all lat_syscall -P 1 stat /var/tmp/lmbench",
        "lmbench_all lat_syscall -P 1 fstat /var/tmp/lmbench",
        "lmbench_all lat_syscall -P 1 open /var/tmp/lmbench",
        "lmbench_all lat_select -n 100 -P 1 file",
        "lmbench_all lat_sig -P 1 install",
        "lmbench_all lat_sig -P 1 catch",
        "lmbench_all lat_sig -P 1 prot lat_sig",
        "lmbench_all lat_pipe -P 1",
        "lmbench_all lat_proc -P 1 fork",
        "lmbench_all lat_proc -P 1 exec",
        "busybox cp hello /tmp",
        "lmbench_all lat_proc -P 1 shell",
        "lmbench_all lmdd label=\"File /var/tmp/XXX write bandwidth:\" of=/var/tmp/XXX move=1m fsync=1 print=3",
        "lmbench_all lat_pagefault -P 1 /var/tmp/XXX",
        "lmbench_all lat_mmap -P 1 512k /var/tmp/XXX",
        "busybox echo file system latency",
        "lmbench_all lat_fs /var/tmp",
        "busybox echo Bandwidth measurements",
        "lmbench_all bw_pipe -P 1",
        "lmbench_all bw_file_rd -P 1 512k io_only /var/tmp/XXX",
        "lmbench_all bw_file_rd -P 1 512k open2close /var/tmp/XXX",
        "lmbench_all bw_mmap_rd -P 1 512k mmap_only /var/tmp/XXX",
        "lmbench_all bw_mmap_rd -P 1 512k open2close /var/tmp/XXX",
        "busybox echo context switch overhead",
        "lmbench_all lat_ctx -P 1 -s 32 2 4 8 16 24 32 64 96",
    ];

    for cmd in cmds.iter() {
        println!("----------------------------------------");
        parse_cmd_and_exec(cmd.to_string());
        println!("----------------------------------------\n");
    }
    println!("------------ lmbench_test finish -------------");
}

fn parse_cmd_and_exec(cmd: String) -> i32 {
    let splited: Vec<_> = cmd.as_str().split('|').collect();
    let process_arguments_list: Vec<_> = splited
        .iter()
        .map(|&cmd| ProcessArguments::new(cmd))
        .collect();
    let mut valid = true;
    for (i, process_args) in process_arguments_list.iter().enumerate() {
        if i == 0 {
            if !process_args.output.is_empty() {
                valid = false;
            }
        } else if i == process_arguments_list.len() - 1 {
            if !process_args.input.is_empty() {
                valid = false;
            }
        } else if !process_args.output.is_empty() || !process_args.input.is_empty() {
            valid = false;
        }
    }
    if process_arguments_list.len() == 1 {
        valid = true;
    }
    if !valid {
        println!("Invalid command: Inputs/Outputs cannot be correctly binded!");
        exit(-4);
    } else {
        // create pipes
        let mut pipes_fd: Vec<[u32; 2]> = Vec::new();
        if !process_arguments_list.is_empty() {
            for _ in 0..process_arguments_list.len() - 1 {
                let mut pipe_fd = [0u32; 2];
                pipe(&mut pipe_fd);
                pipes_fd.push(pipe_fd);
            }
        }
        let mut children: Vec<_> = Vec::new();
        for (i, process_argument) in process_arguments_list.iter().enumerate() {
            let pid = fork();
            if pid == 0 {
                let input = &process_argument.input;
                let output = &process_argument.output;
                let flags = &process_argument.flags;
                let args_copy = &process_argument.args_copy;
                let args_addr = &process_argument.args_addr;
                // redirect input
                if !input.is_empty() {
                    let input_fd = open(input.as_str(), OpenFlags::RDONLY);
                    if input_fd == -1 {
                        println!("Error when opening file {}", input);
                        exit(-4);
                    }
                    let input_fd = input_fd as u32;
                    close(0);
                    assert_eq!(dup(input_fd), 0);
                    close(input_fd);
                }
                // redirect output
                if !output.is_empty() {
                    let output_fd = open(output.as_str(), *flags | OpenFlags::WRONLY);
                    if output_fd == -1 {
                        println!("Error when opening file {}", output);
                        exit(-4);
                    }
                    let output_fd = output_fd as u32;
                    close(1);
                    assert_eq!(dup(output_fd), 1);
                    close(output_fd);
                }
                // receive input from the previous process
                if i > 0 {
                    close(0);
                    let read_end = pipes_fd.get(i - 1).unwrap()[0];
                    assert_eq!(dup(read_end), 0);
                }
                // send output to the next process
                if i < process_arguments_list.len() - 1 {
                    close(1);
                    let write_end = pipes_fd.get(i).unwrap()[1];
                    assert_eq!(dup(write_end), 1);
                }
                // close all pipe ends inherited from the parent process
                for pipe_fd in pipes_fd.iter() {
                    close(pipe_fd[0]);
                    close(pipe_fd[1]);
                }
                // execute new application
                if exec(
                    args_copy[0].as_str(),
                    args_addr.as_slice(),
                    &[core::ptr::null::<u8>()],
                ) == -1
                {
                    println!("Error when executing!");
                    exit(-4);
                }
                unreachable!();
            } else {
                children.push(pid);
            }
        }
        for pipe_fd in pipes_fd.iter() {
            close(pipe_fd[0]);
            close(pipe_fd[1]);
        }
        let mut exit_code: i32 = 0;
        for pid in children.into_iter() {
            let exit_pid = waitpid(pid as usize, &mut exit_code);
            assert_eq!(pid, exit_pid);
            // println!("Shell: Process {} exited with code {}", pid, exit_code);
        }
        exit_code
    }
}

fn test() {
    // autotest();
    lua_test();
    busybox_test();
    // libc_test();
    // lmbench_test();
    println!("!TEST FINISH!");
    // shutdown();
}

#[no_mangle]
pub fn main() -> i32 {
    println!("Rust user shell");
    let mut line: String = String::new();
    test();
    print!("{}", LINE_START);
    loop {
        let c = getchar();
        match c {
            LF | CR => {
                println!("");
                if !line.is_empty() {
                    parse_cmd_and_exec(line.clone());
                    line.clear();
                }
                print!("{}", LINE_START);
            }
            BS | DL => {
                if !line.is_empty() {
                    print!("{}", BS as char);
                    print!(" ");
                    print!("{}", BS as char);
                    line.pop();
                }
            }
            _ => {
                print!("{}", c as char);
                line.push(c as char);
            }
        }
    }
}
